import gzip
import os
import re
import warnings

import ipywidgets as ipw
import numpy as np
import pytest
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.nddata import CCDData
from astropy.utils.data import get_pkg_data_filename
from astropy.utils.exceptions import AstropyWarning
from astropy.wcs import WCS
from astropy.wcs.wcs import FITSFixedWarning

from stellarphot import SourceListData
from stellarphot.gui_tools import comparison_functions as cf
from stellarphot.settings import PhotometryWorkingDirSettings

CCD_SHAPE = [2048, 3073]


def make_ey_uma_image(with_object=False):
    wcs_file = get_pkg_data_filename("../../tests/data/sample_wcs_ey_uma.fits")
    with fits.open(wcs_file) as hdulist:
        with warnings.catch_warnings():
            # Ignore the warning about the WCS having a different number of
            # axes than the (non-existent) image.
            warnings.filterwarnings(
                "ignore",
                message="The WCS transformation has more",
                category=FITSFixedWarning,
            )
            wcs = WCS(hdulist[0].header)
    wcs.pixel_shape = list(reversed(CCD_SHAPE))
    ccd = CCDData(data=np.zeros(CCD_SHAPE), wcs=wcs, unit="adu")
    if with_object:
        ccd.header["object"] = "EY UMa"
    return ccd


def test_comparison_object_creation():
    # This test simply makes sure we can create the object
    comparison_widget = cf.ComparisonViewer()
    assert isinstance(comparison_widget.box, ipw.Box)


def test_fits_file_property():
    # Make sure that the fits file property has the value we expect and that the
    # property is read-only.
    comparison_widget = cf.ComparisonViewer()
    assert comparison_widget.fits_file == comparison_widget._file_chooser
    with pytest.raises(AttributeError):
        comparison_widget.fits_file = None


@pytest.mark.parametrize("source_file_name", [None, "sources.ecsv"])
@pytest.mark.parametrize("has_object", [True, False])
@pytest.mark.remote_data
def test_comparison_properties(tmp_path, has_object, source_file_name):
    # Test that we can load a file...
    ccd = make_ey_uma_image(with_object=has_object)

    file_name = "test.fits"
    ccd.write(tmp_path / file_name, overwrite=True)
    # Change working directory for remainder of test so that the save does not pollute
    # the testing directory.
    os.chdir(tmp_path)

    with warnings.catch_warnings():
        # Ignore the warning about the WCS having non-standard keywords (the SIP
        # distortion parameters).
        warnings.filterwarnings(
            "ignore",
            message="Some non-standard WCS keywords were excluded",
            category=AstropyWarning,
        )
        comparison_widget = cf.ComparisonViewer(
            directory=tmp_path,
            file=str(file_name),
            photom_apertures_file=source_file_name,
        )

    # Check that we have some variables in this field of view, which contains the
    # variable star EY UMa
    assert len(comparison_widget.variables) > 0

    # Check that we have APASS stars too
    table = comparison_widget.generate_table()
    assert "APASS comparison" in table["marker name"]

    # Check that if we show labels then the label names we expect show up in
    # the astrowidgets marker table.
    comparison_widget.show_labels()

    # Suppress a warning about the default marker set containing no stars
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message="Marker set named.*is empty",
            category=UserWarning,
        )
        label_markers = comparison_widget.iw.get_markers(
            marker_name=comparison_widget._label_name
        )
    # There should be a label for each star, so check that the number of labels matches
    # the length of the table of stars excluding the labels.
    table = comparison_widget.generate_table()

    # Drop table entries that are labels
    table = table[table["marker name"] != comparison_widget._label_name]

    assert len(label_markers) == len(table)

    # Make sure the aperture file has been written
    assert os.path.exists(comparison_widget.source_locations.value["source_list_file"])
    # Make sure the aperture file name matches the one we supplied, if
    # we supplied one.
    if source_file_name is not None:
        assert (
            comparison_widget.source_locations.value["source_list_file"]
            == source_file_name
        )

    # Check that a partial photometry settings file exists and that source locations
    # settings have saved correctly.
    partial_settings = PhotometryWorkingDirSettings().load()
    assert (
        partial_settings.source_location_settings.model_dump()
        == comparison_widget.source_locations.value
    )


@pytest.mark.remote_data
def test_loading_second_image_succeeds(tmp_path):
    # Regression test for #384

    # Make a comparison viewer and load EY UMa
    comparison_widget = cf.ComparisonViewer()
    ccd = CCDData.read(
        get_pkg_data_filename(
            "tests/data/TIC_402828941-tiny.fit.bz2", package="stellarphot"
        )
    )
    file_name = "TIC.fits"
    path = tmp_path / file_name
    ccd.write(path, overwrite=True)

    with warnings.catch_warnings():
        # Ignore the warning about the WCS having non-standard keywords (the SIP
        # distortion parameters).
        warnings.filterwarnings(
            "ignore",
            message="Some non-standard WCS keywords were excluded",
            category=AstropyWarning,
        )
        comparison_widget._file_chooser.file_chooser.value = path

    TIC_coord = SkyCoord(305.45862498, 19.43593327, unit="deg")
    # Make sure we have ey uma as the target coordinates
    assert comparison_widget.target_coord.separation(TIC_coord) < 10 * u.arcsec

    # Load a second image
    ccd2 = CCDData.read(get_pkg_data_filename("../../tests/data/wasp-10-tiny.fit.bz2"))
    # ccd2.data = np.zeros(CCD_SHAPE)
    wasp_file = tmp_path / "wasp-10.fits"
    ccd2.write(wasp_file, overwrite=True)
    wasp_coord = SkyCoord(348.99291924, 31.46286002, unit="deg")

    with warnings.catch_warnings():
        # Ignore the warning about the WCS having non-standard keywords (the SIP
        # distortion parameters).
        warnings.filterwarnings(
            "ignore",
            message="Some non-standard WCS keywords were excluded",
            category=AstropyWarning,
        )
        comparison_widget._file_chooser.file_chooser.value = wasp_file
    # Make sure we have WASP-10 as the target coordinates
    assert comparison_widget.target_coord.separation(wasp_coord) < 1 * u.arcsec

    # Also make sure this is where the viewer is actually centered
    # Get the value of the HTML widget that shows the coordinates
    view_coord_text = comparison_widget.iw.children[1].value

    # Extract the coordinates from the text
    match = re.search(
        r"RA: +(\d+:\d+:[.\d]+), +DEC: +([+\d]+:\d+:[.\d]+)", view_coord_text
    )
    ra, dec = match.groups()
    viewer_coord = SkyCoord(ra, dec, unit=(u.hour, u.deg))

    # It is not clear to me what the viewer position defaults to, but it should be
    # close to the target coordinates, where by close I mean "less than the
    # diagonal width of the frame".
    assert viewer_coord.separation(wasp_coord) < 1 * u.degree


@pytest.mark.remote_data
def test_loading_input_source_list(tmp_path):
    # Test that we can load a source list from a file
    comparison_widget = cf.ComparisonViewer()
    compressed_input_source_list = get_pkg_data_filename(
        "tests/data/TIC-402828941-source-list-input.ecsv.gz", package="stellarphot"
    )
    with gzip.open(compressed_input_source_list, "rb") as f:
        source_list_content = f.read()
    input_source_list = tmp_path / "input-sources.ecsv"
    input_source_list.write_text(source_list_content.decode("utf-8"))
    ccd = CCDData.read(
        get_pkg_data_filename(
            "tests/data/TIC_402828941-tiny.fit.bz2", package="stellarphot"
        )
    )
    file_name = "TIC.fits"
    path = tmp_path / file_name
    ccd.write(path, overwrite=True)

    with warnings.catch_warnings():
        # Ignore the warning about the WCS having non-standard keywords (the SIP
        # distortion parameters).
        warnings.filterwarnings(
            "ignore",
            message="Some non-standard WCS keywords were excluded",
            category=AstropyWarning,
        )
        comparison_widget._file_chooser.file_chooser.value = path

    comparison_widget._choose_input_source_list.value = input_source_list

    # Check that the number of tess targets matches the number in the
    # input source list

    comp_viewer_sources = SourceListData.read("source_locations.ecsv")
    input_sources = SourceListData.read(input_source_list)

    num_tess = np.sum(comp_viewer_sources["marker name"] == "TESS Targets")

    assert num_tess == len(input_sources)



from astropy.io.misc.yaml import AstropyDumper, AstropyLoader

from stellarphot.settings import models

__all__ = []


def generate_table_representers(cls):
    """
    Call this method during initialization of a class to add the YAML
    Table representation for the class.
    """
    class_string = f"!{cls.__name__}"

    # Add YAML round-tripping for the model
    def _representer(dumper, model):
        # THIS SHOULD TAP INTO ASTROPY'S YAML DUMPER SOMEHOW
        # This is a little hacky at the moment. It seems like YAML
        # has trouble reading in a dictionary, so though model.model_dump()
        # works fine for writing, we can't construct from the dump.
        #
        # Instead of figuring out the right way to do that, this just dumps
        # the json representation as a string.
        return dumper.represent_mapping(
            class_string, {"model_json_string": model.model_dump_json()}
        )

    def _constructor(loader, node):
        # This loads the simple dictionary we dumped in _representer,
        # then initializes the model with the json string.
        mapping = loader.construct_mapping(node)
        return cls.model_validate_json(mapping["model_json_string"])

    AstropyDumper.add_representer(cls, _representer)
    AstropyLoader.add_constructor(class_string, _constructor)


# This code is deliberately executable so that it can be executed on import, which
# should assure that Table representations are generated for all models.
for model_name in models.__all__:
    model_class = getattr(models, model_name)
    generate_table_representers(model_class)



import ipywidgets as ipw

from stellarphot.gui_tools.comparison_functions import ComparisonViewer
from stellarphot.gui_tools.seeing_profile_functions import SeeingProfileWidget

__all__ = ["ComparisonAndSeeing"]


class ComparisonAndSeeing(ipw.VBox):
    """
    Combined viewer for seeing profile and comparison stars.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.seeing = SeeingProfileWidget()
        self.comparison = ComparisonViewer()
        self.tabs = ipw.Tab(
            children=[self.seeing.box, self.comparison.box],
            titles=["Seeing Profile", "Comparison Stars"],
        )
        self.children = [self.tabs]
        self.comparison.fits_file.file_chooser.observe(
            self._make_observer(
                self.comparison.fits_file.file_chooser,
                self.seeing.fits_file.file_chooser,
            ),
            "_value",
        )
        self.seeing.fits_file.file_chooser.observe(
            self._make_observer(
                self.seeing.fits_file.file_chooser,
                self.comparison.fits_file.file_chooser,
            ),
            "_value",
        )

    @staticmethod
    def _make_observer(source, target):
        def observer(_):
            target.value = source._value

        return observer



import numpy as np
from astropy.coordinates import SkyCoord
from astropy.table import Table
from astropy.utils.data import get_pkg_data_filename

from stellarphot.io import ApertureFileAIJ
from stellarphot.io.aij import ApertureAIJ


def test_aperture_eq():
    ap1 = ApertureAIJ()
    ap2 = ApertureAIJ()
    ap3 = ApertureAIJ()
    ap3.rback1 = 27.0
    assert ap1 == ap2
    assert ap2 == ap3


def test_aperture_file_content():
    # Test that the format of the generated aperture file
    # matches the expected format
    ap = ApertureFileAIJ()
    ref_data = get_pkg_data_filename("data/apertures_as_table.csv")
    ref_table = Table.read(ref_data)

    ap.multiaperture.xapertures = ref_table["x"]

    # AIJ has origin in different place than the reference table.
    ap.multiaperture.yapertures = np.around((4096 - ref_table["y"]), decimals=4)

    ap.multiaperture.raapertures = ref_table["ra"]
    ap.multiaperture.decapertures = ref_table["dec"]

    ap.multiaperture.isrefstar = ref_table["isrefstar"]
    ap.multiaperture.centroidstar = ref_table["centroidstar"]
    ap.multiaperture.isalignstar = ref_table["isalignstar"]

    ap.multiaperture.absmagapertures = ref_table["absmag"]

    ref_aperture_file = get_pkg_data_filename("data/aij-sample-apertures.aperture")

    ref_apertures = ApertureFileAIJ.read(ref_aperture_file)

    assert ref_apertures == ap


def test_aperture_creation_from_table():
    # Check that generating an aperture object from an
    # aperture table gives the right result.

    ref_data = get_pkg_data_filename("data/apertures_as_table.csv")
    ref_table = Table.read(ref_data)

    # Need to create a coord column
    coordinates = SkyCoord(ra=ref_table["ra"], dec=ref_table["dec"], unit="degree")
    ref_table["coord"] = coordinates

    # Delete some columns that are not in the usual aperture table
    del (
        ref_table["ra"],
        ref_table["dec"],
        ref_table["isalignstar"],
        ref_table["centroidstar"],
    )

    # Generate marker names that match those in aperture file
    # Note that the csv reads in the True/False column as text,
    # not as bool.
    ref_table["marker name"] = [
        "APASS comparison" if v == "True" else "TESS target"
        for v in ref_table["isrefstar"]
    ]

    del ref_table["isrefstar"]

    ap_info = ApertureAIJ()
    ap_aij = ApertureFileAIJ.from_table(
        ref_table,
        aperture_rad=ap_info.radius,
        inner_annulus=ap_info.rback1,
        outer_annulus=ap_info.rback2,
    )
    ref_aperture_file = get_pkg_data_filename("data/aij-sample-apertures.aperture")

    ref_apertures = ApertureFileAIJ.read(ref_aperture_file)

    assert ref_apertures == ap_aij



import pandas as pd
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.table import Table

from stellarphot import apass_dr9, vsx_vizier

__all__ = ["read_file", "set_up", "crossmatch_APASS2VSX", "mag_scale", "in_field"]

DESC_STYLE = {"description_width": "initial"}


def read_file(radec_file):
    """
    Read an AIJ radec file with target and/or comparison positions

    Parameters
    ----------

    radec_file : str
        Name of the file

    Returns
    -------

    `astropy.table.Table`
        Table with target information, including a
        `astropy.coordinates.SkyCoord` column.
    """
    df = pd.read_csv(radec_file, names=["RA", "Dec", "a", "b", "Mag"])
    target_table = Table.from_pandas(df)
    ra = target_table["RA"]
    dec = target_table["Dec"]
    target_table["coords"] = SkyCoord(ra=ra, dec=dec, unit=(u.hour, u.degree))
    return target_table


def set_up(ccd):
    """
    Read in sample image and find known variables in the field of view.

    Parameters
    ----------

    ccd: `astropy.nddata.CCDData`
        Sample image.

    Returns
    -------

    vsx: `astropy.table.Table`
        Table with known variables in the field of view.

    """

    try:
        vsx = vsx_vizier(ccd.wcs, radius=0.5 * u.degree)
    except RuntimeError:
        vsx = []
    else:
        ra = vsx["ra"]
        dec = vsx["dec"]
        vsx["coords"] = SkyCoord(ra=ra, dec=dec, unit=u.degree)

    return vsx


def crossmatch_APASS2VSX(CCD, RD, vsx):
    """
    Find APASS stars in FOV and matches APASS stars to VSX and APASS to input targets.

    Parameters
    ----------

    CCD : `astropy.nddata.CCDData`
        Sample image.

    RD : `astropy.table.Table`
        Table with target information, including a
        `astropy.coordinates.SkyCoord` column.

    vsx : `astropy.table.Table`
        Table with known variables in the field of view.

    Returns
    -------

    apass : `stellarphot.CatalogData`
        Table with APASS stars in the field of view.

    v_angle : `astropy.units.Quantity`
        Angular separation between APASS stars and VSX stars.

    RD_angle : `astropy.units.Quantity`
        Angular separation between APASS stars and input targets.
    """
    apass = apass_dr9(CCD.wcs)
    # Use the standard names we have introduced for ra and dec
    ra = apass["ra"]
    dec = apass["dec"]
    # Coordinate units are in degrees now
    apass["coords"] = SkyCoord(ra=ra, dec=dec, unit=u.degree)
    apass_coord = apass["coords"]

    if vsx:
        v_index, v_angle, v_dist = apass_coord.match_to_catalog_sky(vsx["coords"])
    else:
        v_angle = []

    if RD:
        RD_index, RD_angle, RD_dist = apass_coord.match_to_catalog_sky(RD["coords"])
    else:
        RD_angle = []

    return apass, v_angle, RD_angle


def mag_scale(
    cmag, apass, v_angle, RD_angle, brighter_dmag=0.44, dimmer_dmag=0.75, passband="SR"
):
    """
    Select comparison stars that are 1) not close the VSX stars or to other
    target stars and 2) fall within a particular magnitude range.

    Parameters
    ----------

    cmag : float
        Magnitude of the target star.

    apass : `astropy.table.Table`
        Table with APASS stars in the field of view.

    v_angle : `astropy.units.Quantity`
        Angular separation between APASS stars and VSX stars.

    RD_angle : `astropy.units.Quantity`
        Angular separation between APASS stars and input targets.

    brighter_dmag : float, optional
        Maximum difference in magnitude between the target and comparison stars.

    dimmer_dmag : float, optional
        Minimum difference in magnitude between the target and comparison stars.

    passband : str, optional
        Passband to use for selecting the comparison stars.

    Returns
    -------

    apass_good_coord : `astropy.coordinates.SkyCoord`
        Coordinates of the comparison stars.

    good_stars : `astropy.table.Table`
        Table with the comparison stars.
    """
    good_filter = apass["passband"] == passband
    high_mag = apass["mag"] < cmag + dimmer_dmag
    low_mag = apass["mag"] > cmag - brighter_dmag

    too_close = 2.0 * u.arcsec

    if len(v_angle) > 0:
        good_v_angle = v_angle > too_close
    else:
        good_v_angle = True

    if len(RD_angle) > 0:
        good_RD_angle = RD_angle > too_close
    else:
        good_RD_angle = True

    good_stars = good_filter & high_mag & low_mag & good_RD_angle & good_v_angle
    good_apass = apass[good_stars]
    apass_good_coord = good_apass["coords"]
    return apass_good_coord, good_stars


def in_field(apass_good_coord, ccd, apass, good_stars):
    """
    Return APASS stars in the field of view.

    Parameters
    ----------

    apass_good_coord : `astropy.coordinates.SkyCoord`
        Coordinates of the comparison stars.

    ccd : `astropy.nddata.CCDData`
        Sample image.

    apass : `astropy.table.Table`
        Table with APASS stars in the field of view.

    good_stars : `astropy.table.Table`
        Table with the comparison stars.

    Returns
    -------

    ent : `astropy.table.Table`
        Table with APASS stars in the field of view.
    """
    apassx, apassy = ccd.wcs.all_world2pix(apass_good_coord.ra, apass_good_coord.dec, 0)
    ccdx, ccdy = ccd.shape

    xin = (apassx < ccdx) & (0 < apassx)
    yin = (apassy < ccdy) & (0 < apassy)
    xy_in = xin & yin
    apass_good_coord[xy_in]
    nt = apass[good_stars]
    ent = nt[xy_in]
    return ent



import matplotlib.pyplot as plt

__all__ = ["plot_predict_ingress_egress"]


def plot_predict_ingress_egress(
    ingress_time,
    egress_time,
    end_line=1,
    ingress_x_pos=1,
    egress_x_pos=1,
    labels_y_pos=1,
):
    """
    Plot vertical lines at the ingress and egress times and label them.

    Parameters
    ----------
    ingress_time : float
        the beginning of an exoplanet transit

    egress_time : float
        the end of an exoplanet transit

    end_line : float
        offset to move the vertical lines

    ingress_x_pos : float
        offset to center ingress label

    egress_x_pos : float
        offset to center egress label

    labels_y_pos : float
        offset to move ingress and egress labels

    Returns
    -------

    None
        Directly adds lines and labels to the current plot.
    """
    ymin, ymax = plt.ylim()

    # create a vertical line at the ingress time and label it
    plt.vlines(ingress_time, ymin - end_line, ymax, linestyle=(0, (5, 10)), color="red")
    plt.annotate(
        "Predicted Ingress",
        (ingress_time - ingress_x_pos, ymin - labels_y_pos),
        color="red",
    )

    # create a vertical line at the egress time and label it
    plt.vlines(egress_time, ymin - end_line, ymax, linestyle=(0, (5, 10)), color="red")
    plt.annotate(
        "Predicted Egress",
        (egress_time - egress_x_pos, ymin - labels_y_pos),
        fontsize=10,
        color="red",
    )



from astropy.utils.data import get_pkg_data_filename

from stellarphot.settings.fits_opener import FitsOpener


def test_fits_opener_title():
    opener = FitsOpener(title="Choose an image")
    assert opener.file_chooser.title == "Choose an image"


def test_fits_opener_set_value():
    opener = FitsOpener()
    file = get_pkg_data_filename(
        "tests/data/wasp-10-tiny.fit.bz2", package="stellarphot"
    )
    opener.file_chooser.value = file
    assert opener.path.name == "wasp-10-tiny.fit.bz2"



# This file is used to configure the behavior of pytest when using the Astropy
# test infrastructure.
import os

import pytest
from astropy.coordinates import SkyCoord
from astropy.table import Table
from pytest_astropy_header.display import PYTEST_HEADER_MODULES, TESTED_VERSIONS

# astropy-specific-stuff


def pytest_configure(config):
    from astropy.utils.iers import conf as iers_conf

    # Disable IERS auto download for testing
    iers_conf.auto_download = False

    config.option.astropy_header = True
    PYTEST_HEADER_MODULES.pop("h5py", None)

    from .version import version

    packagename = os.path.basename(os.path.dirname(__file__))
    TESTED_VERSIONS[packagename] = version


def pytest_unconfigure():
    from astropy.utils.iers import conf as iers_conf

    # Undo IERS auto download setting for testing
    iers_conf.reset("auto_download")


# stellarphot fixtures


@pytest.fixture
def profile_stars():
    # Make a few round stars
    return Table(
        dict(
            amplitude=[1000, 200, 300],
            x_mean=[30, 100, 150],
            y_mean=[40, 110, 160],
            x_stddev=[4, 4, 4],
            y_stddev=[4, 4, 4],
            theta=[0, 0, 0],
        )
    )


@pytest.fixture
def tess_tic_expected_values():
    # Use this in tests where you need a valid TOI (or TIC) and a coordinate for it.
    return dict(
        tic_id=236158940,
        expected_coords=SkyCoord(ra=313.41953739, dec=34.35164717, unit="degree"),
    )



import os
import warnings
from collections import namedtuple

import ipywidgets as ipw
import matplotlib
import pytest
from astropy.nddata import CCDData
from astrowidgets import ImageWidget
from photutils.datasets import make_noise_image

from stellarphot.gui_tools import (
    seeing_profile_functions as spf,
)
from stellarphot.gui_tools.seeing_profile_functions import (
    AP_SETTING_NEEDS_SAVE,
    AP_SETTING_SAVED,
)
from stellarphot.photometry.tests.fake_image import make_gaussian_sources_image
from stellarphot.photometry.tests.test_profiles import RANDOM_SEED, SHAPE
from stellarphot.settings import (
    Camera,
    Observatory,
    PhotometryApertures,
    PhotometryWorkingDirSettings,
    settings_files,  # This import is needed for mocking
)
from stellarphot.settings.tests.test_models import (
    TEST_CAMERA_VALUES,
)


def test_keybindings():
    def simple_bindmap(bindmap):
        bound_keys = {}
        # The keys of the event map are...messy. This converts them to strings
        for key in bindmap.keys():
            modifier = key[1]
            key_name = key[2]
            bound_keys[str(key[0]) + "".join(modifier) + key_name] = key
        return bound_keys

    # This test assumes the ginga widget backend...
    iw = ImageWidget()
    original_bindings = iw._viewer.get_bindmap().eventmap

    bound_keys = simple_bindmap(original_bindings)
    # Spot check a couple of things before we run our function
    assert "Nonekp_D" in bound_keys
    assert "Nonekp_+" in bound_keys
    assert "Nonekp_left" not in bound_keys

    # rebind
    spf.set_keybindings(iw)
    new_bindings = iw._viewer.get_bindmap().eventmap
    bound_keys = simple_bindmap(new_bindings)
    assert "Nonekp_D" not in bound_keys
    assert "Nonekp_+" in bound_keys
    # Yes, the line below is correct...
    assert new_bindings[bound_keys["Nonekp_left"]]["name"] == "pan_right"


def test_seeing_profile_object_creation():
    # This test simply makes sure we can create the object
    profile_widget = spf.SeeingProfileWidget()
    assert isinstance(profile_widget.box, ipw.Box)


@pytest.fixture(autouse=True)
def fake_settings_dir(mocker, tmp_path):
    # See test_settings_files.py for more information on this fixture.
    # It makes a fake settings directory for each test to use.

    # stellarphot is added to the name of the directory to make sure we start
    # without a stellarphot directory for each test.
    mocker.patch.object(
        settings_files.PlatformDirs, "user_data_dir", tmp_path / "stellarphot"
    )


def test_seeing_profile_properties(tmp_path, profile_stars):
    # Here we make a seeing profile then load an image.
    profile_widget = spf.SeeingProfileWidget(
        camera=Camera(**TEST_CAMERA_VALUES), _testing_path=tmp_path
    )

    # Make a fits file
    image = make_gaussian_sources_image(SHAPE, profile_stars) + make_noise_image(
        SHAPE, mean=10, stddev=100, seed=RANDOM_SEED
    )

    ccd = CCDData(image, unit="adu")
    ccd.header["exposure"] = 30.0
    ccd.header["object"] = "test"
    file_name = tmp_path / "test.fits"
    ccd.write(file_name)

    # Load the file
    profile_widget.fits_file.set_file("test.fits", tmp_path)
    profile_widget.load_fits()

    # Check that a couple of properties are set correctly
    assert profile_widget.object_name == "test"
    assert profile_widget.exposure == 30.0

    # Check that the photometry apertures have defaulted to the
    # default values in the model.
    assert profile_widget.aperture_settings.value == PhotometryApertures().model_dump()

    # Get the event handler that updates plots
    handler = profile_widget._make_show_event()

    # Make a mock event object
    Event = namedtuple("Event", ["data_x", "data_y"])
    star_loc_x, star_loc_y = profile_stars["x_mean"][0], profile_stars["y_mean"][0]
    # Sending a mock event will generate plots that we don't want to see
    # so set the matplotlib backend to a non-interactive one
    matplotlib.use("agg")
    # matplotlib generates a warning that we are using a non-interactive backend
    # so filter that warning out for the remainder of the test. There are at least
    # a couple of times we generate this warning as values are changed.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        handler(profile_widget.iw, Event(star_loc_x, star_loc_y))

        # The FWHM should be close to 9.6
        assert 9 < profile_widget.aperture_settings.value["fwhm"] < 10

        # Get a copy of the current aperture settings
        phot_aps = dict(profile_widget.aperture_settings.value)
        new_radius = phot_aps["radius"] - 2
        # Change the radius by directly setting the value of the widget that holds
        # the value. That ends up being nested fairly deeply...
        profile_widget.aperture_settings.di_widgets["radius"].value = new_radius

        # Make sure the settings are updated
        phot_aps["radius"] = new_radius
        assert profile_widget.aperture_settings.value == phot_aps


def test_seeing_profile_save_apertures(tmp_path):
    # Make sure that saving partial photometery settings works
    os.chdir(tmp_path)

    phot_settings = PhotometryWorkingDirSettings()

    # There should be no saved settings...
    with pytest.raises(ValueError, match="does not exist"):
        phot_settings.load()

    profile_widget = spf.SeeingProfileWidget(
        camera=Camera(**TEST_CAMERA_VALUES), _testing_path=tmp_path
    )

    profile_widget.save()
    assert not profile_widget.aperture_settings.savebuttonbar.unsaved_changes

    settings = phot_settings.load()
    assert settings.camera == Camera(**TEST_CAMERA_VALUES)
    assert settings.photometry_apertures == PhotometryApertures(
        radius=1, annulus_width=1, gap=1
    )


def test_seeing_profile_save_box_title(tmp_path):
    # Save box title ends with different characters depending on whether values
    # need to be saved.

    # Change working directory since settings file is saved there.
    os.chdir(tmp_path)

    profile_widget = spf.SeeingProfileWidget(
        camera=Camera(**TEST_CAMERA_VALUES), _testing_path=tmp_path
    )
    profile_widget.aperture_settings.di_widgets["radius"].value = 3

    assert profile_widget.aperture_settings.savebuttonbar.unsaved_changes
    assert AP_SETTING_NEEDS_SAVE in profile_widget.ap_title.value

    profile_widget.save()
    assert AP_SETTING_SAVED in profile_widget.ap_title.value


def test_seeing_profile_error_messages_no_star(tmp_path):
    # Make sure the appropriate error message is displayed when a click happens on
    # a region with no star, and that the message only appears once.
    profile_widget = spf.SeeingProfileWidget(
        camera=Camera(**TEST_CAMERA_VALUES), _testing_path=tmp_path
    )
    # Make a fits file with no stars
    image = make_noise_image(SHAPE, mean=10, stddev=1, seed=RANDOM_SEED)

    ccd = CCDData(image, unit="adu")
    ccd.header["exposure"] = 30.0
    ccd.header["object"] = "test"
    file_name = tmp_path / "test.fits"
    ccd.write(file_name)

    # Load the file
    profile_widget.fits_file.set_file("test.fits", tmp_path)
    profile_widget.load_fits()

    # Get the event handler that updates plots
    handler = profile_widget._make_show_event()

    # Make a mock event object
    Event = namedtuple("Event", ["data_x", "data_y"])
    star_loc_x, star_loc_y = SHAPE[0] // 2, SHAPE[1] // 2
    # Sending a mock event will generate plots that we don't want to see
    # so set the matplotlib backend to a non-interactive one
    matplotlib.use("agg")
    # matplotlib generates a warning that we are using a non-interactive backend
    # so filter that warning out for the remainder of the test. There are at least
    # a couple of times we generate this warning as values are changed.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        assert len(profile_widget.error_console.outputs) == 0

        # Clicking once should generate an error...
        handler(profile_widget.iw, Event(star_loc_x, star_loc_y))
        assert len(profile_widget.error_console.outputs) == 1
        assert (
            "No star found at this location"
            in profile_widget.error_console.outputs[0]["data"]["text/plain"]
        )

        # Clicking a second time should also just have one error
        handler(profile_widget.iw, Event(star_loc_x, star_loc_y))
        assert len(profile_widget.error_console.outputs) == 1


def test_seeing_profile_no_observatory():
    # This test checks that with no observatory set, there is no TESS
    # related box displayed.
    profile_widget = spf.SeeingProfileWidget()
    assert profile_widget.save_toggle is None
    assert profile_widget.tess_box.layout.visibility == "hidden"


@pytest.mark.parametrize("tess_code", [None, "dummy"])
def test_seeing_profile_with_observatory(tess_code):
    # Test two cases here:
    #  1. The observatory has TESS_telescope_code set to a string
    #  2. The observatory has TESS_telescope_code set to None
    observatory = Observatory(
        name="test",
        latitude=0.0,
        longitude=0.0,
        elevation="0.0 m",
        TESS_telescope_code=tess_code,
    )
    profile_widget = spf.SeeingProfileWidget(observatory=observatory)
    if tess_code is None:
        assert profile_widget.save_toggle is None
    else:
        assert isinstance(profile_widget.save_toggle, ipw.ToggleButton)



from stellarphot.transit_fitting.gui import (
    exotic_settings_widget,
    get_values_from_widget,
)


def test_json_has_stuff():
    widget = exotic_settings_widget()
    stuff = get_values_from_widget(widget, key="known")
    assert "user_info" in stuff.keys()



import functools
import logging

import ipyautoui
import ipywidgets as ipw
import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.coordinates.name_resolve import NameResolveError
from astropy.table import Table
from ipyautoui.custom import FileChooser

try:
    from astrowidgets import ImageWidget
except ImportError:
    from astrowidgets.ginga import ImageWidget

from stellarphot import SourceListData
from stellarphot.gui_tools.seeing_profile_functions import set_keybindings
from stellarphot.settings import (
    PartialPhotometrySettings,
    PhotometryWorkingDirSettings,
    SourceLocationSettings,
    ui_generator,
)
from stellarphot.settings.custom_widgets import SettingWithTitle, Spinner
from stellarphot.settings.fits_opener import FitsOpener
from stellarphot.utils.comparison_utils import (
    crossmatch_APASS2VSX,
    in_field,
    mag_scale,
    set_up,
)

__all__ = ["make_markers", "wrap", "ComparisonViewer"]

DESC_STYLE = {"description_width": "initial"}


def make_markers(iw, RD, vsx, ent, name_or_coord=None):
    """
    Add markers for APASS, TESS targets, VSX.  Also center on object/coordinate.

    Parameters
    ----------

    iw : `astrowidgets.ImageWidget`
        Ginga widget.

    RD : `astropy.table.Table`
        Table with target information, including a
        `astropy.coordinates.SkyCoord` column.

    vsx : `astropy.table.Table`
        Table with known variables in the field of view.

    ent : `astropy.table.Table`
        Table with APASS stars in the field of view.

    name_or_coord : str or `astropy.coordinates.SkyCoord`, optional
        Name or coordinates of the target.

    Returns
    -------

    None
        Markers are added to the image in Ginga widget.
    """
    iw.zoom_level = "fit"

    try:
        iw.reset_markers()
    except AttributeError:
        iw.remove_all_markers()

    if RD:
        iw.marker = {"type": "circle", "color": "green", "radius": 10}
        iw.add_markers(
            RD, skycoord_colname="coords", use_skycoord=True, marker_name="TESS Targets"
        )

    if name_or_coord is not None:
        if isinstance(name_or_coord, str):
            iw.center_on(SkyCoord.from_name(name_or_coord))
        else:
            iw.center_on(name_or_coord)

    if vsx:
        # TODO: if astrowidgets ever gets more sensible marker types use those instead
        # of using the internal interface to marker
        iw._marker = functools.partial(
            iw.dc.Box,
            xradius=10,
            yradius=10,
            color="blue",
        )
        iw.add_markers(
            vsx, skycoord_colname="coords", use_skycoord=True, marker_name="VSX"
        )
    # TODO: if astrowidgets ever gets more sensible marker types use those instead
    # of using the internal interface to marker
    iw._marker = functools.partial(
        iw.dc.Triangle,
        xradius=10,
        yradius=10,
        color="red",
    )
    iw.add_markers(
        ent,
        skycoord_colname="coords",
        use_skycoord=True,
        marker_name="APASS comparison",
    )
    iw.marker = {"type": "cross", "color": "red", "radius": 6}


def wrap(imagewidget, outputwidget):
    """
    Utility function to let you click to select/deselect comparisons.

    Parameters
    ----------

    imagewidget : `astrowidgets.ImageWidget`
        Ginga widget.

    outputwidget : `ipywidgets.Output`
        Output widget for printing information.

    """

    def cb(viewer, event, data_x, data_y):  # noqa: ARG001
        """
        The signature of this function must have the four arguments above.
        """
        i = imagewidget._viewer.get_image()

        try:
            imagewidget.next_elim += 1
        except AttributeError:
            imagewidget.next_elim = 1

        ra, dec = i.wcs.wcs.all_pix2world(event.data_x, event.data_y, 0)
        out_skycoord = SkyCoord(ra=ra, dec=dec, unit=(u.degree, u.degree))

        try:
            all_table = imagewidget.get_markers(marker_name="all")
        except AttributeError:
            all_table = imagewidget.get_all_markers()

        with outputwidget:
            index, d2d, d3d = out_skycoord.match_to_catalog_sky(all_table["coord"])
            if d2d < 10 * u.arcsec:
                mouse = all_table["coord"][index].separation(all_table["coord"])
                rat = mouse < 1 * u.arcsec
                elims = [
                    name
                    for name in all_table["marker name"][rat]
                    if name.startswith("elim")
                ]
                if not elims:
                    imagewidget.add_markers(
                        all_table[rat],
                        skycoord_colname="coord",
                        use_skycoord=True,
                        marker_name=f"elim{imagewidget.next_elim}",
                    )
                else:
                    for elim in elims:
                        try:
                            imagewidget.remove_markers_by_name(marker_name=elim)
                        except AttributeError:
                            imagewidget.remove_markers(marker_name=elim)

            else:
                print("sorry try again")
                imagewidget._viewer.onscreen_message("Click closer to a star")

    return cb


class ComparisonViewer:
    """
    A class to store an instance of the comparison viewer.

    Parameters
    ----------

    file : str, optional
        File to open. Defaults to "".

    directory : str, optional
        Directory to open file from. Defaults to '.'.

    target_mag : float, optional
        Magnitude of the target. Defaults to 10.

    bright_mag_limit : float, optional
        Bright magnitude limit for APASS stars. Defaults to 8.

    dim_mag_limit : float, optional
        Dim magnitude limit for APASS stars.  Defaults to 17.

    targets_from_file : str, optional
        File with target information.  Defaults to None.

    object_coordinate : `astropy.coordinates.SkyCoord`, optional
        Coordinates of the target. Defaults to None.

    photom_apertures_file : str, optional
        File to save photometry aperture information to.  Defaults to None.

    overwrite_outputs: bool, optional
        Whether to overwrite existing output files. Defaults to True.

    observatory : `stellarphot.settings.Observatory`, optional
        Observatory information.

    Attributes
    ----------

    photom_apertures_file : str
        File to save photometry aperture information to.

    box : `ipywidgets.Box`
        Box containing the widgets.

    bright_mag_limit : float
        Bright magnitude limit for APASS stars.

    dim_mag_limit : float
        Dim magnitude limit for APASS stars.

    iw : `ginga.util.grc.RemoteClient`
        Ginga widget.

    overwrite_outputs: bool
        Whether to overwrite existing output files. Defaults to True.

    target_coord : `astropy.coordinates.SkyCoord`
        Coordinates of the target.

    targets_from_file : str
        File with target information.

    target_mag : float
        Magnitude of the target.

    variables : `astropy.table.Table`
    """

    def __init__(
        self,
        file="",
        directory=".",
        target_mag=10,
        bright_mag_limit=8,
        dim_mag_limit=15,
        targets_from_file=None,
        object_coordinate=None,
        photom_apertures_file=None,
        overwrite_outputs=True,
        observatory=None,
    ):
        self._label_name = "labels"
        self._circle_name = "target circle"
        self._file_chooser = FitsOpener()

        self._directory = directory
        self.target_mag = target_mag
        self.bright_mag_limit = bright_mag_limit
        self.dim_mag_limit = dim_mag_limit
        self.targets_from_file = targets_from_file
        self.target_coord = object_coordinate
        self.observatory = observatory

        # This function defines several attributes in addition to returning the box and
        # image viewer. You should take a look at it to see what it does.
        self.box, self.iw = self._viewer()

        self.photometry_settings = PhotometryWorkingDirSettings()

        if photom_apertures_file is not None:
            original_settings = self.source_locations.value.copy()
            original_settings["source_list_file"] = photom_apertures_file
            self.source_locations.value = original_settings

        self.overwrite_outputs = overwrite_outputs

        if file:
            self._file_chooser.set_file(file, directory=directory)
            self._set_file(None)

        # Save the source location settings
        self.source_locations.savebuttonbar.bn_save.click()

        self._make_observers()

    def _init(self):
        """
        Handles aspects of initialization that need to be deferred until
        a file is chosen.
        """
        # Read in the ccd data here and load the image into viewer instead of doing
        # it behind the scenes in set_up and make_markers
        self.ccd = self.fits_file.ccd
        self.fits_file.load_in_image_widget(self.iw)

        spinner = Spinner(message="Loading variable/comparison stars")
        legend = self._legend_spinner_box.children[0]
        spinner.start()
        self._legend_spinner_box.children = [spinner]
        self.help_stuff.selected_index = 1

        self.vsx = set_up(self.ccd)

        apass, vsx_apass_angle, targets_apass_angle = crossmatch_APASS2VSX(
            self.ccd, self.targets_from_file, self.vsx
        )

        self._legend_spinner_box.children = [legend]

        apass_good_coord, good_stars = mag_scale(
            self.target_mag,
            apass,
            vsx_apass_angle,
            targets_apass_angle,
            brighter_dmag=self.target_mag - self.bright_mag_limit,
            dimmer_dmag=self.dim_mag_limit - self.target_mag,
        )

        apass_comps = in_field(apass_good_coord, self.ccd, apass, good_stars)

        # Set the object here so that the viewer is properly centered
        self._set_object()

        make_markers(
            self.iw,
            self.targets_from_file,
            self.vsx,
            apass_comps,
            name_or_coord=self.target_coord,
        )

    @property
    def fits_file(self):
        return self._file_chooser

    @property
    def photom_apertures_file(self):
        return self.source_locations.value["source_list_file"]

    @property
    def variables(self):
        """
        An `astropy.table.Table` of the variables in the class.
        """
        comp_table = self.generate_table()
        new_vsx_mark = comp_table["marker name"] == "VSX"
        idx, _, _ = comp_table["coord"][new_vsx_mark].match_to_catalog_sky(
            self.vsx["coords"]
        )
        our_vsx = self.vsx[idx]
        our_vsx["star_id"] = comp_table["star_id"][new_vsx_mark]
        return our_vsx

    def _set_object(self):
        """
        Try to automatically set object name immediately after file is chosen.
        """
        try:
            self.object_name.value = (
                f"<h2> Object: {self._file_chooser.header['object']}</h2>"
            )
            self._object = self._file_chooser.header["object"]
        except KeyError:
            # No object, will show placeholder
            self.object_name.value = "<h2>Object unknown</h2>"
            self._object = ""

        # We have a name, try to get coordinates from it
        try:
            self.target_coord = SkyCoord.from_name(self._object)
        except NameResolveError:
            # If there are no coordinations so far then use the center of the frame. The
            # generation of  the source list table depends on the target coordinates.
            self.target_coord = self.ccd.wcs.pixel_to_world(
                self.ccd.shape[0] / 2, self.ccd.shape[1] / 2
            )

    def _save_source_location_file(self):
        """
        Save the source location file.
        """
        # Save the initial source list if there is a file name for it and if the target
        # coordinates are known. The target coordinates are needed to generate the
        # source list table.
        if (
            self.source_locations.value["source_list_file"] is not None
            and self.target_coord is not None
        ):
            self._save_aperture_to_file(None)

    def _set_file(self, change):  # noqa: ARG002
        """
        Widget callbacks need to accept a change argument, even if not used.
        """
        self._init()
        self._save_source_location_file()

    def _set_target_file(self, change):
        """
        Set the target file.
        """
        input_source_list_name = change["new"]
        self.targets_from_file = SourceListData.read(input_source_list_name)
        # Only call init if a file has been chosen
        if self._file_chooser.file_chooser.value != ".":
            self._init()
            self._save_source_location_file()

    def _make_observers(self):
        self._show_labels_button.observe(self._show_label_button_handler, names="value")
        self._save_var_info.on_click(self._save_variables_to_file)
        self._file_chooser.file_chooser.observe(self._set_file, names="_value")
        self._choose_input_source_list.observe(self._set_target_file, names="_value")

    def _save_variables_to_file(self, button=None, filename=""):  # noqa: ARG002
        """
        Widget button callbacks need to be able to take an argument. It is called
        button above the the button will be passed as the first positional argument.
        """
        if not filename:
            filename = "variables.csv"
        # Export variables as CSV (overwrite existing file if it exists)
        try:
            self.variables.write(filename, overwrite=self.overwrite_outputs)
        except OSError as err:
            raise OSError(
                f"Existing file ({filename}) can not be overwritten. "
                "Set overwrite_outputs=True to address this."
            ) from err

    def _show_label_button_handler(self, change):
        value = change["new"]
        if value:
            # Showing labels can take a bit, disable while in progress
            self._show_labels_button.disabled = True
            self.show_labels()
            self._show_labels_button.disabled = False
        else:
            self.remove_labels()
        self._show_labels_button.description = self._show_labels_button.descriptions[
            value
        ]

    def _save_aperture_to_file(self, button=None, filename=""):  # noqa: ARG002
        """
        Widget button callbacks need to be able to take an argument. It is called
        button above the the button will be passed as the first positional argument.
        """
        if not filename:
            filename = self.photom_apertures_file

        # Convert aperture table into a SourceList objects and output
        targets_table = self.generate_table()
        # Assign units to columns
        targets_table["ra"] = targets_table["ra"] * u.deg
        targets_table["dec"] = targets_table["dec"] * u.deg
        targets_table["x"] = targets_table["x"] * u.pixel
        targets_table["y"] = targets_table["y"] * u.pixel
        # Drop redundant sky position column
        targets_table.remove_columns(["coord"])
        # Build sources list
        targets2sourcelist = {"x": "xcenter", "y": "ycenter"}
        sources = SourceListData(
            input_data=targets_table, colname_map=targets2sourcelist
        )

        # Export aperture file as CSV (overwrite existing file if it exists)
        try:
            sources.write(filename, overwrite=self.overwrite_outputs)
        except OSError as err:
            raise OSError(
                f"Existing file ({filename}) can not be overwritten."
                "Set overwrite_outputs=True to address this."
            ) from err

    def _make_control_bar(self):
        self._show_labels_button = ipw.ToggleButton(description="Click to show labels")
        self._show_labels_button.descriptions = {
            True: "Click to hide labels",
            False: "Click to show labels",
        }

        self._save_var_info = ipw.Button(description="Save variable info")

        controls = ipw.HBox(
            children=[
                self._show_labels_button,
                self._save_var_info,
            ]
        )

        return controls

    def save(self):
        """
        Save all of the settings we have to a partial settings file.
        """
        self.photometry_settings.save(
            PartialPhotometrySettings(
                source_location_settings=self.source_locations.value,
            ),
            update=True,
        )

        # For some reason the value of unsaved_changes is not updated until after this
        # function executes, so we force its value here.
        self.source_locations.savebuttonbar.unsaved_changes = False
        # Update the save box title to reflect the save
        self.source_and_title.decorate_title()

    def _viewer(self):
        header = ipw.HTML(
            value="""
        <h3>Click and drag or use arrow keys to pan, use +/- keys to zoom.</h3>
        <h3>Shift-left click (or Crtl-left click) to exclude star as target
        or comp. Click again to include.</h3>
        """
        )

        legend = ipw.HTML(
            value="""
        <ul>
        <li>Green circles -- Gaia stars within 2.5 arcmin of target</li>
        <li>Red triangles -- Comparison stars from APASS</li>
        <li>Blue squares -- VSX variables</li>
        <li>Red × -- Exclude as target or comp</li>
        </ul>
        """
        )

        iw = ImageWidget()
        out = ipw.Output()
        set_keybindings(iw)
        bind_map = iw._viewer.get_bindmap()
        gvc = iw._viewer.get_canvas()
        bind_map.map_event(None, ("shift",), "ms_left", "cursor")
        gvc.add_callback("cursor-down", wrap(iw, out))

        self.object_name = ipw.HTML(value="<h2>Object: </h2>")
        self._object = None
        controls = self._make_control_bar()

        # Capture the logging message issued about setting the file chooser to a file
        # that does not exist.
        original_logging_level = ipyautoui.custom.filechooser.logger.level
        ipyautoui.custom.filechooser.logger.setLevel(logging.CRITICAL)

        # The logging message will be generated here if it is generated.
        self.source_locations = ui_generator(
            SourceLocationSettings, max_field_width="75px"
        )
        ipyautoui.custom.filechooser.logger.setLevel(original_logging_level)

        self.source_and_title = SettingWithTitle(
            "Source location settings", self.source_locations
        )

        self.source_locations.savebuttonbar.fns_onsave_add_action(self.save)

        # Put the legend in a box whose children can be changed to a
        # spinner while loading.
        self._legend_spinner_box = ipw.VBox()
        self._legend_spinner_box.children = [legend]

        self.help_stuff = ipw.Accordion(children=[header, self._legend_spinner_box])
        self.help_stuff.titles = ["Help", "Legend"]
        box = ipw.VBox()
        inner_box = ipw.HBox()
        source_legend_box = ipw.VBox()
        source_legend_box.children = [
            self.help_stuff,
            self.object_name,
            self.source_and_title,
        ]
        inner_box.children = [iw, source_legend_box]

        # Add a file chooser for an input source list
        self._choose_input_source_list = FileChooser(
            filter_pattern="*.ecsv",
        )
        self._choose_input_source_list.title = "OPTIONAL: Choose input source list"

        box.children = [
            self._file_chooser.file_chooser,
            self._choose_input_source_list,
            inner_box,
            controls,
        ]

        return box, iw

    def generate_table(self):
        """
        Generate the table of stars to use for the aperture file.

        Returns
        -------
        comp_table : `astropy.table.Table`
            Table of stars to use for the aperture file.
        """
        try:
            all_table = self.iw.get_all_markers()
        except AttributeError:
            all_table = self.iw.get_markers(marker_name="all")

        elims = np.array([name.startswith("elim") for name in all_table["marker name"]])
        elim_table = all_table[elims]
        comp_table = all_table[~elims]

        index, d2d, d3d = elim_table["coord"].match_to_catalog_sky(comp_table["coord"])
        comp_table.remove_rows(index)

        # Add separate RA and Dec columns for ease in processing later
        comp_table["ra"] = comp_table["coord"].ra.degree
        comp_table["dec"] = comp_table["coord"].dec.degree

        # Calculate how far each is from target
        comp_table["separation"] = self.target_coord.separation(comp_table["coord"])

        # Add dummy column for sorting in the order we want
        comp_table["sort"] = np.zeros(len(comp_table))

        # Set sort order
        apass_mark = comp_table["marker name"] == "APASS comparison"
        vsx_mark = comp_table["marker name"] == "VSX"
        tess_mark = (comp_table["marker name"] == "TESS Targets") | (
            comp_table["separation"] < 0.3 * u.arcsec
        )

        comp_table["sort"][apass_mark] = 2
        comp_table["sort"][vsx_mark] = 1
        comp_table["sort"][tess_mark] = 0

        # Sort the table
        comp_table.sort(["sort", "separation"])

        # Assign the IDs
        comp_table["star_id"] = range(1, len(comp_table) + 1)

        return comp_table

    def show_labels(self):
        """
        Show the labels for the stars.

        Returns
        -------

        None
            Labels for the stars are shown.
        """
        plot_names = []
        comp_table = self.generate_table()
        label_size = 15
        original_mark = self.iw._marker
        for star in comp_table:
            star_id = star["star_id"]
            if star["marker name"] == "TESS Targets":
                label = f"T{star_id}"
                self.iw._marker = functools.partial(
                    self.iw.dc.Text,
                    text=label,
                    fontsize=label_size,
                    fontscale=False,
                    color="green",
                )
                self.iw.add_markers(
                    Table(data=[[star["x"] + 20], [star["y"] - 20]], names=["x", "y"]),
                    marker_name=self._label_name,
                )

            elif star["marker name"] == "APASS comparison":
                label = f"C{star_id}"
                self.iw._marker = functools.partial(
                    self.iw.dc.Text,
                    text=label,
                    fontsize=label_size,
                    fontscale=False,
                    color="red",
                )
                self.iw.add_markers(
                    Table(data=[[star["x"] + 20], [star["y"] - 20]], names=["x", "y"]),
                    marker_name=self._label_name,
                )

            elif star["marker name"] == "VSX":
                label = f"V{star_id}"
                self.iw._marker = functools.partial(
                    self.iw.dc.Text,
                    text=label,
                    fontsize=label_size,
                    fontscale=False,
                    color="blue",
                )
                self.iw.add_markers(
                    Table(data=[[star["x"] + 20], [star["y"] - 20]], names=["x", "y"]),
                    marker_name=self._label_name,
                )
            else:
                label = f"U{star_id}"
                print(f"Unrecognized marker name: {star['marker name']}")
            plot_names.append(label)
        self.iw._marker = original_mark

    def remove_labels(self):
        """
        Remove the labels for the stars.

        Returns
        -------

        None
            Labels for the stars are removed.
        """
        try:
            try:
                self.iw.remove_markers(marker_name=self._label_name)
            except AttributeError:
                self.iw.remove_markers_by_name(marker_name=self._label_name)
        except ValueError:
            # No labels, keep going
            pass

    def show_circle(self, radius=2.5 * u.arcmin, pixel_scale=0.56 * u.arcsec / u.pixel):
        """
        Show a circle around the target.

        Parameters
        ----------

        radius : `astropy.units.Quantity`, optional
            Radius of circle. The default is ``2.5*u.arcmin``.

        pixel_scale : `astropy.units.Quantity`, optional
            Pixel scale of image. The default is ``0.56*u.arcsec/u.pixel``.

        Returns
        -------

        None
            Circle is shown around the target.
        """
        radius_pixels = np.round((radius / pixel_scale).to(u.pixel).value, decimals=0)
        orig_marker = self.iw.marker
        self.iw.marker = {"color": "yellow", "radius": radius_pixels, "type": "circle"}
        self.iw.add_markers(
            Table(data=[[self.target_coord]], names=["coords"]),
            skycoord_colname="coords",
            use_skycoord=True,
            marker_name=self._circle_name,
        )
        self.iw.marker = orig_marker

    def remove_circle(self):
        """
        Remove the circle around the target.

        Returns
        -------

        None
            Circle is removed from the image.
        """
        try:
            self.iw.remove_markers(marker_name=self._circle_name)
        except AttributeError:
            self.iw.remove_markers_by_name(marker_name=self._circle_name)

    def tess_field_view(self):
        """
        Show the whole TESS field of view including circle around target,
        but hide labels.

        Returns
        -------

        None
            Shows image as described above.
        """

        # Show whole field of view
        self.iw.zoom_level = "fit"

        # Show the circle
        self.show_circle()

        # Turn off labels -- too cluttered
        self.remove_labels()

    def tess_field_zoom_view(self, width=6 * u.arcmin):
        """
        Zoom in on the TESS field of view.

        Parameters
        ----------

        width : `astropy.units.Quantity`, optional
            Width of field of view. The default is ``6*u.arcmin``.

        Returns
        -------

        None
            Zooms in on the image as described above.
        """

        # Turn off labels -- too cluttered
        self.remove_labels()

        left_side = self.ccd.wcs.pixel_to_world(0, self.ccd.shape[1] / 2)
        right_side = self.ccd.wcs.pixel_to_world(
            self.ccd.shape[0], self.ccd.shape[1] / 2
        )
        fov = left_side.separation(right_side)

        view_ratio = width / fov
        # fit first to get zoom level at full field of view
        self.iw.zoom_level = "fit"

        # Then set it to what we actually want...
        self.iw.zoom_level = self.iw.zoom_level / view_ratio

        # Show the circle
        self.show_circle()



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os

from astropy.tests.runner import TestRunner

__all__ = ["__version__", "test"]

try:
    from .version import version as __version__
except ImportError:
    __version__ = ""

# Create the test function for self test
test = TestRunner.make_test_runner_in(os.path.dirname(__file__))
test.__test__ = False



from astroquery.mast import Catalogs

__all__ = ["get_tic_info"]


def get_tic_info(TIC_ID):
    """
    Get the information about this TIC ID from the TESS Input Catalog
    at MAST.

    Parameters
    ----------

    TIC_ID : int
        9 or 10 digit TIC ID number.

    Returns
    -------

    `astropy.table.Table`
        Astropy table withinformation about the TIC object.

    """
    catalog_data = Catalogs.query_criteria(catalog="Tic", ID=TIC_ID)
    return catalog_data



# Objects that contains the user settings for the program.

import re
from copy import deepcopy
from pathlib import Path
from typing import Annotated, Any, Literal, Optional, TypeVar

from astropy.coordinates import EarthLocation, Latitude, Longitude, SkyCoord
from astropy.time import Time
from astropy.units import Quantity, Unit, UnitConversionError
from astropy.utils import lazyproperty
from pydantic import (
    AfterValidator,
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    NonNegativeFloat,
    PositiveFloat,
    PositiveInt,
    create_model,
    field_validator,
    model_validator,
)

from .aavso_models import AAVSOFilters
from .astropy_pydantic import (
    AstropyValidator,
    EquivalentTo,
    QuantityType,
    UnitType,
    WithPhysicalType,
    _UnitQuantTypePydanticAnnotation,
)

__all__ = [
    "Camera",
    "LoggingSettings",
    "PartialPhotometrySettings",
    "PassbandMap",
    "PhotometryApertures",
    "PhotometryFileSettings",
    "PhotometryRunSettings",
    "PhotometrySettings",
    "PhotometryOptionalSettings",
    "Exoplanet",
    "Observatory",
    "SourceLocationSettings",
]

# Most models should use the default configuration, but it can be customized if needed.
MODEL_DEFAULT_CONFIGURATION = ConfigDict(
    # Make sure default values are valid
    validate_default=True,
    # Make sure changes to values made after initialization are valid
    validate_assignment=True,
    # Make sure there are no extra fields
    extra="forbid",
)


# Make a type for a non-empty string for use in name fields
# Names may not have any leading or trailing spaces, and cannot simply
# be spaces. Though this could be implemented as a regular expression
# pattern, that leads to a validation message likely to confuse users,
# along the lines of (for pattern r"^\S$|^\S.*\S$"):
#
# String should match pattern "^\\S$|^\\S.*\\S$"
#
# Instead, the custom validation function below checks for errors and
# raises a ValueError with a more user-friendly message.
def _non_empty_string_validator(value):
    if not value.strip():
        raise ValueError("name must not be empty or contain only whitespace.")

    if not re.search(r"^\S$|^\S.*\S$", value):
        # Name must have exactly one non-whitespace character or
        # at least two non-whitespace characters with any amount of
        # other characters between them.
        raise ValueError("name must not have leading or trailing whitespace.")

    return value


NonEmptyStr = Annotated[str, AfterValidator(_non_empty_string_validator)]


def _extract_short_description(docstring: str) -> str:
    """
    Extract the first line of the docstring as a short description.
    """
    # Everything up the the first blank line is the short description
    short_desc = docstring.split("\n\n")[0]

    # This may have some extra whitespace at the beginning and end, remove that
    short_desc = short_desc.strip()

    # split the string and rejoin with spaces in case there are multiple spaces
    # between words, e.g. caused by a multi-line docstring
    return " ".join(short_desc.split())


def add_degree_to_float(value, _handler):
    """
    Translate a value that can be a number to a string with "degree" appended.
    """
    try:
        as_number = float(value)
    except (ValueError, TypeError):
        # A value error will happen if the value is not a number.
        # A type error will happen at least in the case where the value is
        # an astropy Quantity.
        return value
    else:
        return f"{as_number} degree"


class BaseModelWithTableRep(BaseModel):
    """
    Class to add to a pydantic model YAML serialization to an Astropy table.
    """

    # NOTE WELL that this will set the configuration for all subclasses of this
    model_config = MODEL_DEFAULT_CONFIGURATION

    def __init__(self, *arg, **kwargs):
        super().__init__(*arg, **kwargs)


class Camera(BaseModelWithTableRep):
    """
    A class to represent a CCD-based camera.

    Parameters
    ----------

    name : str
        The name of the camera; can be anything that helps the user identify
        the camera.

    data_unit : `astropy.units.Unit`
        The unit of the data.

    gain : `astropy.units.Quantity`
        The gain of the camera in units such the product of `gain`
        times the image data has units equal to that of the `read_noise`.

    read_noise : `astropy.units.Quantity`
        The read noise of the camera with units.

    dark_current : `astropy.units.Quantity`
        The dark current of the camera in units such that, when multiplied by
        exposure time, the unit matches the units of the `read_noise`.

    pixel_scale : `astropy.units.Quantity`
        The pixel scale of the camera in units of arcseconds per pixel.

    max_data_value : `astropy.units.Quantity`
        The maximum pixel value to allow while performing photometry. Pixel values
        above this will be set to ``NaN``. The unit must be ``data_unit``.

    Attributes
    ----------

    data_unit : `astropy.units.Unit`
        The unit of the data.

    gain : `astropy.units.Quantity`
        The gain of the camera in units such the product of `gain`
        times the image data has units equal to that of the `read_noise`.

    name : str
        The name of the camera; can be anything that helps the user identify
        the camera.

    read_noise : `astropy.units.Quantity`
        The read noise of the camera with units.

    dark_current : `astropy.units.Quantity`
        The dark current of the camera in units such that, when multiplied by
        exposure time, the unit matches the units of the `read_noise`.

    pixel_scale : `astropy.units.Quantity`
        The pixel scale of the camera in units of arcseconds per pixel.

    max_data_value : `astropy.units.Quantity`
        The maximum pixel value to allow while performing photometry. Pixel values
        above this will be set to ``NaN``. The unit must be ``data_unit``.

    Notes
    -----
    The gain, read noise, and dark current are all assumed to be constant
    across the entire CCD.

    Examples
    --------
    >>> from astropy import units as u
    >>> from stellarphot.settings import Camera
    >>> camera = Camera(data_unit="adu",
    ...                 gain=1.0 * u.electron / u.adu,
    ...                 name="test camera",
    ...                 read_noise=1.0 * u.electron,
    ...                 dark_current=0.01 * u.electron / u.second,
    ...                 pixel_scale=0.563 * u.arcsec / u.pixel,
    ...                 max_data_value=50000 * u.adu)
    >>> camera.data_unit
    Unit("adu")
    >>> camera.gain
    <Quantity 1. electron / adu>
    >>> camera.name
    'test camera'
    >>> camera.read_noise
    <Quantity 1. electron>
    >>> camera.dark_current
    <Quantity 0.01 electron / s>
    >>> camera.pixel_scale
    <Quantity 0.563 arcsec / pix>
    >>> camera.max_data_value
    <Quantity 50000. adu>
    """

    # This ensures that just the first line of the docstring is used as the
    # model description. The json schema is really hard to read if the
    # description is too long.

    # WHen a subclass has a ConfigDict it is merged with the parent class's
    # ConfigDict.
    model_config = ConfigDict(
        json_schema_extra=dict(
            description=_extract_short_description(__doc__),
        )
    )

    name: Annotated[
        NonEmptyStr,
        Field(
            description="Name of the camera",
            examples=["SBIG FakeCam", "ZWO NadaCam", "CG16m"],
        ),
    ]
    data_unit: UnitType = Field(
        description="units of the data", examples=["adu", "DN", "count"]
    )
    gain: QuantityType = Field(
        description="unit should be consistent with data and read noise",
        examples=["1.5 electron / adu", "1.0 electron / DN", "1.0 photon / count"],
    )
    read_noise: QuantityType = Field(
        description="unit should be consistent with dark current",
        examples=["10.0 electron", "10.0 electron", "10.0 photon"],
    )
    dark_current: QuantityType = Field(
        description="unit consistent with read noise, per unit time",
        examples=[
            "0.01 electron / second",
            "0.01 electron / second",
            "0.01 photon / second",
        ],
    )
    pixel_scale: Annotated[
        QuantityType,
        EquivalentTo(Unit("arcsec / pix")),
        Field(description="units of angle per pixel", examples=["0.6 arcsec / pix"]),
    ]
    max_data_value: Annotated[
        QuantityType,
        Field(
            description="maximum data value while performing photometry",
            examples=["50000 adu", "50000 DN", "50000 count"],
            gt=0,
        ),
    ]

    # Run the model validator after the default validator. Unlike in pydantic 1,
    # mode="after" passes in an instance as an argument not a value.
    @model_validator(mode="after")
    def validate_gain(self):
        # Get read noise units
        rn_unit = Quantity(self.read_noise).unit
        # Get dark current units
        dark_unit = Quantity(self.dark_current).unit

        # Check that gain and read noise have compatible units, that is that
        # gain is read noise per data unit.
        gain = self.gain

        try:
            gain.to(self.read_noise.unit / self.data_unit)
        except UnitConversionError as e:
            raise ValueError(
                f"Gain units {gain.unit} are not compatible with "
                f"read noise units {rn_unit}."
            ) from e

        # Check that dark current and read noise have compatible units, that is
        # that dark current is read noise per second.
        dark_current = self.dark_current
        try:
            dark_current.to(self.read_noise.unit / Unit("s"))
        except UnitConversionError as e:
            raise ValueError(
                f"Dark current units {dark_unit} are not "
                f"compatible with read noise units {rn_unit}."
            ) from e

        # Check that maximum data value is consistent with data units
        if self.max_data_value.unit != self.data_unit:
            raise ValueError(
                f"Maximum data value units {self.max_data_value.unit} "
                f"are not consistent with data units {self.data_unit}."
            )
        return self


class PhotometryApertures(BaseModelWithTableRep):
    """
    Settings for aperture photometry.

    Parameters
    ----------

    radius : int
        Radius of the aperture in pixels, must be greater than or equal to 1.

    gap : int
        Distance between the radius and the inner annulus in pixels, must be greater
        than or equal to 1.

    annulus_width : int
        Width of the annulus in pixels, must be greater than or equal to 1.

    fwhm : float
        Full width at half maximum of the typical star in pixels.

    Attributes
    ----------

    inner_annulus : int
        Radius of the inner annulus in pixels.

    outer_annulus : int
        Radius of the outer annulus in pixels.

    Examples
    --------

    To create an `PhotometryApertures` object, you can pass in the radius, gap,
    and annulus_width as keyword arguments:

    >>> aperture_settings = PhotometryApertures(
    ...     radius=4,
    ...     gap=10,
    ...     annulus_width=15,
    ...     fwhm=3.0
    ... )
    """

    # This ensures that just the first line of the docstring is used as the
    # model description. The json schema is really hard to read if the
    # description is too long.

    # WHen a subclass has a ConfigDict it is merged with the parent class's
    # ConfigDict.
    model_config = ConfigDict(
        json_schema_extra=dict(
            description=_extract_short_description(__doc__),
        )
    )

    radius: Annotated[
        PositiveInt,
        Field(
            default=1,
            description="radius of circular aperture, in pixels",
            json_schema_extra=dict(autoui="ipywidgets.BoundedIntText"),
        ),
    ]
    gap: Annotated[
        PositiveInt,
        Field(
            default=1,
            description="Size of gap between aperture and annulus, in pixels",
            json_schema_extra=dict(autoui="ipywidgets.BoundedIntText"),
        ),
    ]
    annulus_width: Annotated[
        PositiveInt,
        Field(
            default=1,
            description=("distance between inner and outer radii of annulus in pixels"),
            json_schema_extra=dict(autoui="ipywidgets.BoundedIntText"),
        ),
    ]
    # Disable the UI element by default because it is often calculate from an image
    fwhm: Annotated[PositiveFloat, Field(disabled=True, default=1.0, title="FWHM")]

    @property
    def inner_annulus(self):
        """
        Radius of the inner annulus in pixels.
        """
        return self.radius + self.gap

    @property
    def outer_annulus(self):
        """
        Radius of the outer annulus in pixels.
        """
        return self.inner_annulus + self.annulus_width


class PhotometryFileSettings(BaseModelWithTableRep):
    """
    An evolutionary step on the way to having a monolithic set of photometry settings.
    """

    image_folder: Path = Field(
        show_only_dirs=True,
        default="",
        description="Folder containing the calibrated images",
    )
    aperture_settings_file: Path = Field(filter_pattern="*.json", default="")
    aperture_locations_file: Path = Field(
        filter_pattern=["*.ecsv", "*.csv"], default=""
    )


class Observatory(BaseModelWithTableRep):
    """
    Class to represent an observatory.

    Parameters
    ----------
    name : str
        Name of the observatory.

    latitude : `astropy.coordinates.Latitude` or other valid latitude representation
        Latitude of the observatory. Use a positive number for north and negative
        for south.

    longitude : `astropy.coordinates.Longitude` or other valid longitude representation
        Longitude of the observatory. Use a positive number for east and negative
        for west.

    elevation : `astropy.units.Quantity`
        Elevation of the observatory.

    AAVSO_code : str, optional
        AAVSO observer code.

    TESS_telescope_code : str, optional
        TESS telescope code.

    Examples
    --------

    The first example shows how to set up an observatory whose latitude and longitude
    are 30° North and 100° West, respectively, and is at an elevation of 1000 meters.

    >>> from astropy.coordinates import Latitude, Longitude
    >>> from astropy import units as u
    >>> from stellarphot.settings import Observatory
    >>> observatory = Observatory(
    ...     name="test observatory",
    ...     latitude=Latitude(30.0 * u.deg),
    ...     longitude=Longitude(-100.0 * u.deg),
    ...     elevation=1000 * u.m,
    ... )
    >>> observatory
    Observatory(name='test observatory', latitude=<Latitude 30. deg>,
    longitude=<Longitude 260. deg>, elevation=<Quantity 1000. m>, AAVSO_code=None,
    TESS_telescope_code=None)

    Note that units can be omitted from the latitude and longitude if and only if the
    units are degrees and the values are in decimal degrees.

    >>> # You can also just provide numbers for the latitude and longitude
    >>> observatory = Observatory(
    ...     name="test observatory",
    ...     latitude=30.0,
    ...     longitude=-100.0,
    ...     elevation=1000 * u.m,
    ... )
    >>> observatory
    Observatory(name='test observatory', latitude=<Latitude 30. deg>,
    longitude=<Longitude 260. deg>, elevation=<Quantity 1000. m>, AAVSO_code=None,
    TESS_telescope_code=None)

    An observatory located at 46° 00' 00.00" South and 96° 00' 00.00" East, and at an
    elevation of 2300 meters would be created like this:

    >>> observatory = Observatory(
    ...     name="test observatory",
    ...     latitude="-46d00m00.00s",
    ...     longitude="96d00m00.00s",
    ...     elevation=2.3 * u.km,
    ... )
    >>> observatory
    Observatory(name='test observatory', latitude=<Latitude -46. deg>,...


    It would be fine to use decmial degrees for the latitude and longitude in the
    above example, but the example is given in part to show how to use sexagesimal
    notation.

    """

    # This ensures that just the first line of the docstring is used as the
    # model description. The json schema is really hard to read if the
    # description is too long.

    # WHen a subclass has a ConfigDict it is merged with the parent class's
    # ConfigDict.
    model_config = ConfigDict(
        json_schema_extra=dict(
            description=_extract_short_description(__doc__),
        )
    )

    name: Annotated[
        NonEmptyStr,
        Field(description="Name of the observatory", examples=["My Observatory"]),
    ]
    latitude: Annotated[
        Latitude,
        _UnitQuantTypePydanticAnnotation,
        BeforeValidator(add_degree_to_float),
        Field(
            description="Latitude of the observatory",
            examples=[
                "46.8738",
                "46.8738 degree",
                "46:52:25.68 degree",
                "46d52m25.68s",
            ],
        ),
    ]
    longitude: Annotated[
        Longitude,
        _UnitQuantTypePydanticAnnotation,
        BeforeValidator(add_degree_to_float),
        Field(
            description="Longitude of the observatory",
            examples=[
                "-96.7678",
                "-96d46m04.08s",
                "263.2322 degree",
                "263d13m55.92s",
            ],
        ),
    ]
    elevation: Annotated[
        QuantityType,
        WithPhysicalType("length"),
        Field(
            description="Elevation of the observatory",
            examples=["1000 m", "1 km", "3.241e-14 pc", "1e12 nm"],
        ),
    ]
    AAVSO_code: Annotated[str | None, Field(description="AAVSO code for observer")] = (
        None
    )
    TESS_telescope_code: Annotated[
        str | None, Field(description="Code for observatory used by TFOP SG01")
    ] = None

    @lazyproperty
    def earth_location(self):
        """
        Return an `astropy.coordinates.EarthLocation` object for the observatory.
        """
        return EarthLocation(
            lat=self.latitude, lon=self.longitude, height=self.elevation
        )


class SourceLocationSettings(BaseModelWithTableRep):
    """
    Settings for the location of the source list and the image files.

    Parameters
    ----------
    source_list_file : str
        Name of a file with a table of extracted sources with positions in terms of
        pixel coordinates OR RA/Dec coordinates. If both positions provided,
        the one that will be used is determined by `use_coordinates`. For RA/Dec
        coordinates to be used, `ccd_image` must have a valid WCS.

    use_coordinates : `typing.Literal["sky", "pixel"]`, optional
        If ``'pixel'``, use the x/y positions in the sourcelist for
        performing aperture photometry.  If ``'sky'``, use the ra/dec
        positions in the sourcelist and the WCS of the `ccd_image` to
        compute the x/y positions on the image.

    shift_tolerance : `pydantic.NonNegativeFloat`, optional
        Since source positions need to be computed on each image using
        the sky position and WCS, the computed x/y positions are refined
        afterward by centroiding the sources.  This setting controls
        the tolerance in pixels for the shift between the the computed
        positions and the refined positions, in pixels.  The expected
        shift shift should not be more than the FWHM, so a measured FWHM
        might be a good value to provide here.

    Examples
    --------
    >>> from stellarphot.settings import SourceLocationSettings
    >>> source_location_settings = SourceLocationSettings(
    ...     source_list_file="source_list.ecsv",
    ...     use_coordinates="sky",
    ...     shift_tolerance=5.0
    ... )
    >>> source_location_settings
    SourceLocationSettings(source_list_file='source_list.ecsv', use_coordinates='sky',
    shift_tolerance=5.0)
    """

    # This ensures that just the first line of the docstring is used as the
    # model description. The json schema is really hard to read if the
    # description is too long.

    # WHen a subclass has a ConfigDict it is merged with the parent class's
    # ConfigDict.
    model_config = ConfigDict(
        json_schema_extra=dict(
            description=_extract_short_description(__doc__),
        )
    )

    source_list_file: Annotated[
        str,
        Field(
            json_schema_extra=dict(
                autoui="ipyautoui.custom.FileChooser",
                filter_pattern=["*.ecsv", "*.csv"],
            )
        ),
    ] = "source_locations.ecsv"
    use_coordinates: Annotated[
        Literal["sky", "pixel"],
        Field(description="coordinates to use for locating sources in image"),
    ] = "sky"

    shift_tolerance: Annotated[
        NonNegativeFloat,
        Field(
            description=(
                "Maximum shift between source position in list and "
                "in image, in pixels"
            )
        ),
    ] = 5.0


class PhotometryOptionalSettings(BaseModelWithTableRep):
    """
    Options for performing photometry.

    Parameters
    ----------
    include_dig_noise : bool, optional (Default: True)
        If ``True``, include the digitization noise in the calculation of the
        noise for each observation.  If ``False``, only the Poisson noise from
        the source and the sky will be included.

    reject_too_close : bool, optional (Default: True)
        If ``True``, any sources that are closer than twice the aperture radius
        are rejected.  If ``False``, all sources in field are used.

    reject_background_outliers : bool, optional (Default: True)
        If ``True``, sigma clip the pixels in the annulus to reject outlying
        pixels (e.g. like stars in the annulus)

    fwhm_by_fit : bool, optional (default: True)
        If ``True``, the FWHM will be calculated by fitting a Gaussian to
        the star. If ``False``, the FWHM will be calculated by finding the
        second order moments of the light distribution. Default is ``True``.

    method : `typing.Literal["exact", "center", "subpixel"]`, optional
        How to handle partial pixels in the aperture.  If ``'exact'``, the fraction of
        the flux included is the fraction of the pixel within the aperture. If
        ``'center'``, whether a pixel's flux is included is determined by whether the
        center of the pixel is within the aperture. If ``'subpixel'``, the flux included
        is determineid by breaking  region into subpixels. The default is ``exact``.
        For more information, see the
        `photutils documentation <https://photutils.readthedocs.io/en/stable/aperture.html#aperture-and-pixel-overlap>`_.

    Examples
    --------

    In many cases the default options are fine:

    >>> from stellarphot.settings import PhotometryOptionalSettings
    >>> photometry_options = PhotometryOptionalSettings()
    >>> photometry_options
    PhotometryOptionalSettings(include_dig_noise=True, reject_too_close=True,...

    You can also set options explicitly when you create the options:

    >>> photometry_options = PhotometryOptionalSettings(
    ...     include_dig_noise=True,
    ...     reject_too_close=False,
    ...     reject_background_outliers=True,
    ...     fwhm_by_fit=True,
    ...     method="center"
    ... )
    >>> photometry_options
    PhotometryOptionalSettings(include_dig_noise=True, reject_too_close=False,...
    reject_background_outliers=True, fwhm_by_fit=True, method='center')

    You can also change individual options after the object is created:

    >>> photometry_options.reject_background_outliers = False
    >>> photometry_options.reject_background_outliers
    False
    """

    # This ensures that just the first line of the docstring is used as the
    # model description. The json schema is really hard to read if the
    # description is too long.

    # WHen a subclass has a ConfigDict it is merged with the parent class's
    # ConfigDict.
    model_config = ConfigDict(
        json_schema_extra=dict(
            description=_extract_short_description(__doc__),
        )
    )

    include_dig_noise: Annotated[
        bool,
        Field(
            title="Include Digitization Noise",
            description=(
                "Should the digitization noise be included in the calculation of "
                "the noise for each observation?"
            ),
        ),
    ] = True

    reject_too_close: Annotated[
        bool,
        Field(
            description=(
                "Should sources that are too close each other be "
                "excluded from photometry?"
            )
        ),
    ] = True

    reject_background_outliers: Annotated[
        bool,
        Field(
            description=(
                "Should extreme pixels in the annulus be rejected "
                "(e.g. from a star in the annulus)?"
            )
        ),
    ] = True

    fwhm_by_fit: Annotated[
        bool,
        Field(
            description=(
                "Should the FWHM be calculated by fitting a Gaussian to "
                "the star or from image moments?"
            )
        ),
    ] = True

    method: Annotated[
        Literal["exact", "center", "subpixel"],
        Field(
            description="How to handle partial pixels in the aperture.",
        ),
    ] = "exact"


class PassbandMapEntry(BaseModel):
    """
    A mapping from a single filter name to its corresponding AAVSO filter name.

    Parameters
    ----------
    your_filter_name : str
        Instrumental filter name.

    aavso_filter_name : `stellarphot.settings.AAVSOFilters`
        AAVSO filter name.

    """

    # This ensures that just the first line of the docstring is used as the
    # model description. The json schema is really hard to read if the
    # description is too long.
    model_config = ConfigDict(
        json_schema_extra=dict(
            description=_extract_short_description(__doc__),
        )
    )

    your_filter_name: Annotated[
        NonEmptyStr, Field(description="Instrumental Filter Name")
    ]
    aavso_filter_name: Annotated[AAVSOFilters, Field(title="AAVSO Filter Name")]


class PassbandMap(BaseModelWithTableRep):
    """
    Class to represent a mapping from one set of filter names to another that behaves
    like a `dict`.

    Parameters
    ----------

    name : str
        Name of the passband map.

    your_filter_names_to_aavso : list[`stellarphot.settings.PassbandMapEntry`]
        A list of pairs of your filter name and the corresponding AAVSO filter name.
        This is used to rename the passband entries in the output photometry table.
        Note that, as shown in the example below, you can initialize this with a
        dictionary, and it will be converted to a list of `PassbandMapEntry` objects.

    Notes
    -----

    This class behaves like a dictionary in terms of accessing individual entries but
    you _cannot_ use the `dict` methods to modify the object. This means, for example,
    that if ``my_map`` is a `PassbandMap` object, you can access the AAVSO passband
    that corresponds to your ``B`` passband with ``my_map["B"]`` but you _cannot_
    set entries like this ``my_map["B"] = "B"`` and you _cannot_ delete entries like
    this ``del my_map["B"]``.

    Examples
    --------
    >>> from stellarphot.settings import PassbandMap
    >>> passband_map = PassbandMap(
    ...     name="Sample map",
    ...     your_filter_names_to_aavso={"B": "B", "rp": "SR"}
    ... )
    >>> passband_map
    PassbandMap(name='Sample map', your_filter_names_to_aavso=[PassbandMapEntry(your_...
    >>> # You can access the AAVSO filter name for a given filter name using dict syntax
    >>> passband_map["B"]
    'B'
    >>> passband_map["rp"]
    'SR'
    >>> # If you prefer you can access the individual entries in
    >>> # the list of PassbandMapEntry
    >>> passband_map.your_filter_names_to_aavso[1]
    PassbandMapEntry(your_filter_name='rp', aavso_filter_name=<AAVSOFilters.SR: 'SR'>)
    >>> # Getting the AAVSO filter name this way is a little cumbersome though:
    >>> passband_map.your_filter_names_to_aavso[1].aavso_filter_name.value
    'SR'

    """

    # This ensures that just the first line of the docstring is used as the
    # model description. The json schema is really hard to read if the
    # description is too long.

    # WHen a subclass has a ConfigDict it is merged with the parent class's
    # ConfigDict.
    model_config = ConfigDict(
        json_schema_extra=dict(
            description=_extract_short_description(__doc__),
        )
    )

    name: Annotated[
        NonEmptyStr,
        Field(
            description="Name of the passband map",
            examples=["Filter wheel 1"],
        ),
    ]
    your_filter_names_to_aavso: list[PassbandMapEntry]

    def model_post_init(self, __context: Any) -> None:
        # Create a dictionary from the list of entries so that the object
        # can behave like a dictionary.
        self._dict = {
            entry.your_filter_name: entry.aavso_filter_name.value
            for entry in self.your_filter_names_to_aavso
        }

    @field_validator("your_filter_names_to_aavso", mode="before")
    @classmethod
    def validate_your_filter_names_to_aavso(cls, v):
        if isinstance(v, PassbandMap):
            return v.your_filter_names_to_aavso
        elif isinstance(v, dict):
            return [
                PassbandMapEntry(your_filter_name=k, aavso_filter_name=v)
                for k, v in v.items()
            ]
        else:
            return v

    # All of the remaining methods are to make the object behave like a dictionary.
    # It would have been preferable to subclass `collections.UserDict` but that
    # doesn't work with pydantic models because UserDict objects have a .data attribute
    # but we don't want a pydantic field named "data".
    def items(self):
        return self._dict.items()

    def keys(self):
        return self._dict.keys()

    def values(self):
        return self._dict.values()

    def get(self, key, default=None):
        return self._dict.get(key, default)

    def __contains__(self, key):
        return key in self._dict

    def __getitem__(self, key):
        return self._dict[key]

    def __iter__(self):
        return iter(self._dict)


class LoggingSettings(BaseModelWithTableRep):
    """
    Settings for logging.

    Parameters
    ----------
    logfile : str, optional (Default: None)
        Name of the file to which log messages should be written.  It will
        be created in the `directory_with_images` directory.  If None,
        no log file will be created.

    console_log: bool, optional (Default: True)
        If ``True``, log messages will be written to stdout.  If ``False``, log
        messages will not be written to stdout.

    Examples
    --------
    >>> from stellarphot.settings import LoggingSettings
    >>> logging_settings = LoggingSettings()
    >>> logging_settings
    LoggingSettings(logfile=None, console_log=True)
    """

    # This ensures that just the first line of the docstring is used as the
    # model description. The json schema is really hard to read if the
    # description is too long.

    # WHen a subclass has a ConfigDict it is merged with the parent class's
    # ConfigDict.
    model_config = ConfigDict(
        json_schema_extra=dict(
            description=_extract_short_description(__doc__),
        )
    )

    logfile: Annotated[str | None, Field(description="File to save log to")] = None
    console_log: Annotated[bool, Field(description="Show log on console?")] = True


SCHEMA_EXTRAS = dict(show_null=True)


class PhotometrySettings(BaseModelWithTableRep):
    """
    Settings for performing aperture photometry.

    Parameters
    ----------

    camera : `stellarphot.settings.Camera`
        Camera object which has gain, read noise and dark current set.

    observatory : `stellarphot.settings.Observatory`
        Observatory information.  Used for calculating the BJD.

    photometry_apertures : `stellarphot.settings.PhotometryApertures`
        Radius, inner and outer annulus radii settings and FWHM.

    source_locations : `stellarphot.settings.SourceLocationSettings`
        Settings for the location of the sources for which photometry
        will be performed. See the documentation for
        `~stellarphot.settings.SourceLocationSettings` for details.

    photometry_options : `stellarphot.settings.PhotometryOptions`
        Several options for the details of performing the photometry. See the
        documentation for `~stellarphot.settings.PhotometryOptions` for details.

    passband_map: `stellarphot.settings.PassbandMap`, optional
        A dictionary containing instrumental passband names as keys and
        AAVSO passband names as values. This is used to rename the passband
        entries in the output photometry table from what is in the source list
        to be AAVSO standard names, if available for that filter.

    logging_settings : `stellarphot.settings.LoggingSettings`
        Settings for logging. See the documentation for
        `~stellarphot.settings.LoggingSettings` for details.

    """

    # This ensures that just the first line of the docstring is used as the
    # model description. The json schema is really hard to read if the
    # description is too long.

    # WHen a subclass has a ConfigDict it is merged with the parent class's
    # ConfigDict.
    model_config = ConfigDict(
        json_schema_extra=dict(
            description=_extract_short_description(__doc__),
        )
    )

    camera: Annotated[
        Camera,
        Field(
            description=_extract_short_description(Camera.__doc__),
            json_schema_extra=SCHEMA_EXTRAS,
        ),
    ]
    observatory: Annotated[
        Observatory,
        Field(
            description=_extract_short_description(Observatory.__doc__),
            json_schema_extra=SCHEMA_EXTRAS,
        ),
    ]
    passband_map: Annotated[
        PassbandMap | None,
        Field(
            description=_extract_short_description(PassbandMap.__doc__),
            json_schema_extra=SCHEMA_EXTRAS,
        ),
    ]
    photometry_apertures: Annotated[
        PhotometryApertures,
        Field(
            description=_extract_short_description(PhotometryApertures.__doc__),
            json_schema_extra=SCHEMA_EXTRAS,
        ),
    ]
    source_location_settings: Annotated[
        SourceLocationSettings,
        Field(
            description=_extract_short_description(SourceLocationSettings.__doc__),
            json_schema_extra=SCHEMA_EXTRAS,
        ),
    ]
    photometry_optional_settings: Annotated[
        PhotometryOptionalSettings,
        Field(
            description=_extract_short_description(PhotometryOptionalSettings.__doc__),
            json_schema_extra=SCHEMA_EXTRAS,
        ),
    ]
    logging_settings: Annotated[
        LoggingSettings,
        Field(
            description=_extract_short_description(LoggingSettings.__doc__),
            json_schema_extra=SCHEMA_EXTRAS,
        ),
    ]


# The code for _make_partial_model is adapted from
# https://github.com/pydantic/pydantic/issues/3120#issuecomment-1528030416

BaseModelT = TypeVar("BaseModelT", bound=BaseModel)


def _make_partial_model(model: type[BaseModelT], default=None) -> type[BaseModelT]:
    new_fields = {}

    for field_name, field_info in model.model_fields.items():
        new = deepcopy(field_info)
        new.default = default
        new.annotation = Optional[field_info.annotation]  # type: ignore  # noqa: UP007
        new_fields[field_name] = (new.annotation, new)
    return create_model(  # type: ignore
        f"Partial{model.__name__}",
        __base__=model,
        __doc__=model.__doc__,
        __module__=model.__module__,
        **new_fields,
    )


PartialPhotometrySettings = _make_partial_model(PhotometrySettings)


class Exoplanet(BaseModelWithTableRep):
    """
    Create an object representing an Exoplanet.

    Parameters
    ----------

    epoch : `astropy.time.Time`, optional
        Epoch of the exoplanet.

    period : `astropy.units.Quantity`, optional
        Period of the exoplanet.

    Identifier : str
        Identifier of the exoplanet.

    coordinate : `astropy.coordinates.SkyCoord`
        Coordinates of the exoplanet.

    depth : float
        Depth of the exoplanet.

    duration : `astropy.units.Quantity`, optional
        Duration of the exoplanet transit.

    Examples
    --------

    To create an `Exoplanet` object, you can pass in the epoch,
     period, identifier, coordinate, depth, and duration as keyword arguments:

    >>> from astropy.time import Time
    >>> from astropy.coordinates import SkyCoord
    >>> from astropy import units as u
    >>> planet  = Exoplanet(epoch=Time(2455909.29280, format="jd"),
    ...                     period=1.21749 * u.day,
    ...                     identifier="KELT-1b",
    ...                     coordinate=SkyCoord(ra="00:01:26.9169",
    ...                                         dec="+39:23:01.7821",
    ...                                         frame="icrs",
    ...                                         unit=("hour", "degree")),
    ...                     depth=0.006,
    ...                     duration=120 * u.min)
    """

    model_config = MODEL_DEFAULT_CONFIGURATION

    epoch: Annotated[Time, AstropyValidator] | None = None
    period: Annotated[QuantityType, WithPhysicalType("time")] | None = None
    identifier: str
    coordinate: Annotated[SkyCoord, AstropyValidator]
    depth: float | None = None
    duration: Annotated[QuantityType, WithPhysicalType("time")] | None = None


class PhotometryRunSettings(BaseModel):
    """
    Settings for a photometry run.

    Parameters
    ----------
    photometry_settings : `stellarphot.settings.PhotometrySettings`
        Settings for performing aperture photometry.

    exoplanet : `stellarphot.settings.Exoplanet`, optional
        Information about an exoplanet transit.

    """

    directory_with_images: Path = "."
    photometry_settings_file: Path = "photometry_settings.json"
    reject_unmatched: Annotated[
        bool, Field(description="Blah blah", tooltip="WOMA")
    ] = True
    object_of_interest: str | None = None



import warnings

import ipywidgets as ipw
import numpy as np
from astropy.io import fits
from astropy.table import Table

try:
    from astrowidgets import ImageWidget
except ImportError:
    from astrowidgets.ginga import ImageWidget

import matplotlib.pyplot as plt

from stellarphot.io import TessSubmission
from stellarphot.photometry import CenterAndProfile
from stellarphot.photometry.photometry import EXPOSURE_KEYWORDS
from stellarphot.plotting import seeing_plot
from stellarphot.settings import (
    PartialPhotometrySettings,
    PhotometryApertures,
    PhotometryWorkingDirSettings,
    SavedSettings,
    ui_generator,
)
from stellarphot.settings.custom_widgets import ChooseOrMakeNew
from stellarphot.settings.fits_opener import FitsOpener

__all__ = [
    "set_keybindings",
    "SeeingProfileWidget",
]

DESC_STYLE = {"description_width": "initial"}
AP_SETTING_NEEDS_SAVE = "❗️"
AP_SETTING_SAVED = "✅"
DEFAULT_SAVE_TITLE = "Save aperture and camera"


# TODO: maybe move this into SeeingProfileWidget unless we anticipate
# other widgets using this.
def set_keybindings(image_widget, scroll_zoom=False):
    """
    Set image widget keyboard bindings. The bindings are:

    + Pan by click-and-drag or with arrow keys.
    + Zoom by scrolling or using the ``+``/``-`` keys.
    + Adjust contrast by Ctrl-right click and drag
    + Reset contrast with shift-right-click.

    Any existing key bindings are removed.

    Parameters
    ----------

    image_widget : `astrowidgets.ImageWidget`
        Image widget on which to set the key bindings.

    scroll_zoom : bool, optional
        If True, zooming can be done by scrolling the mouse wheel.
        Default is False.

    Returns
    -------

    None
        Adds key bindings to the image widget.
    """
    bind_map = image_widget._viewer.get_bindmap()
    # Displays the event map...
    # bind_map.eventmap
    bind_map.clear_event_map()
    bind_map.map_event(None, (), "ms_left", "pan")
    if scroll_zoom:
        bind_map.map_event(None, (), "pa_pan", "zoom")

    # bind_map.map_event(None, (), 'ms_left', 'cursor')
    # contrast with right mouse
    bind_map.map_event(None, (), "ms_right", "contrast")

    # shift-right mouse to reset contrast
    bind_map.map_event(None, ("shift",), "ms_right", "contrast_restore")
    bind_map.map_event(None, ("ctrl",), "ms_left", "cursor")

    # Bind +/- to zoom in/out
    bind_map.map_event(None, (), "kp_+", "zoom_in")
    bind_map.map_event(None, (), "kp_=", "zoom_in")
    bind_map.map_event(None, (), "kp_-", "zoom_out")
    bind_map.map_event(None, (), "kp__", "zoom_out")

    # Bind arrow keys to panning
    # There is NOT a typo below. I want the keys to move the image in the
    # direction of the arrow
    bind_map.map_event(None, (), "kp_left", "pan_right")
    bind_map.map_event(None, (), "kp_right", "pan_left")
    bind_map.map_event(None, (), "kp_up", "pan_down")
    bind_map.map_event(None, (), "kp_down", "pan_up")


class SeeingProfileWidget:
    """
    A class for storing an instance of a widget displaying the seeing profile
    of stars in an image.

    Parameters
    ----------
    imagewidget : `astrowidgets.ImageWidget`, optional
        ImageWidget instance to use for the seeing profile.

    width : int, optional
        Width of the seeing profile widget. Default is 500 pixels.

    camera : `stellarphot.settings.Camera`, optional
        Camera instance to use for calculating the signal to noise ratio. If
        ``None``, the signal to noise ratio will not be calculated.

    observatory : `stellarphot.settings.Observatory`, optional
        Observatory instance to use for setting the TESS telescope information.
        If `None`, or if the `~stellarphot.settings.Observatory.TESS_telescope_code`
        is `None`, the TESS settings will not be displayed.

    Attributes
    ----------

    ap_t : `ipywidgets.IntText`
        Text box for the aperture radius.

    box : `ipywidgets.VBox`
        Box containing the seeing profile widget.

    container : `ipywidgets.VBox`
        Container for the seeing profile widget.

    object_name : str
        Name of the object in the FITS file.

    in_t : `ipywidgets.IntText`
        Text box for the inner annulus.

    iw : `astrowidgets.ImageWidget`
        ImageWidget instance used for the seeing profile.

    object_name : str
        Name of the object in the FITS file.

    out : `ipywidgets.Output`
        Output widget for the seeing profile.

    out2 : `ipywidgets.Output`
        Output widget for the integrated counts.

    out3 : `ipywidgets.Output`
        Output widget for the SNR.

    out_t : `ipywidgets.IntText`
        Text box for the outer annulus.

    rad_prof : `RadialProfile`
        Radial profile of the star.

    save_aps : `ipywidgets.Button`
        Button to save the aperture settings.

    tess_box : `ipywidgets.VBox`
        Box containing the TESS settings.

    """

    def __init__(
        self,
        imagewidget=None,
        width=500,
        camera=None,
        observatory=None,
        _testing_path=None,
    ):
        if not imagewidget:
            imagewidget = ImageWidget(
                image_width=width, image_height=width, use_opencv=True
            )

        self.photometry_settings = PhotometryWorkingDirSettings()
        self.iw = imagewidget

        self.observatory = observatory

        # If a camera is provided make sure it has already been saved.
        # If it has not been saved, raise an error.
        if camera is not None:
            saved = SavedSettings(_testing_path=_testing_path)
            if camera not in saved.cameras.as_dict.values():
                saved.add_item(camera)

        # Do some set up of the ImageWidget
        set_keybindings(self.iw, scroll_zoom=False)
        bind_map = self.iw._viewer.get_bindmap()
        bind_map.map_event(None, ("shift",), "ms_left", "cursor")
        gvc = self.iw._viewer.get_canvas()
        self._mse = self._make_show_event()
        gvc.add_callback("cursor-down", self._mse)

        # Outputs to hold the graphs
        self.seeing_profile_plot = ipw.Output()
        self.curve_growth_plot = ipw.Output()
        self.snr_plot = ipw.Output()

        # Include an error console to display messages to the user
        self.error_console = ipw.Output()

        # Build the larger widget
        self.container = ipw.VBox()
        self.fits_file = FitsOpener(title=self._format_title("Choose an image"))
        self.camera_chooser = ChooseOrMakeNew(
            "camera", details_hideable=True, _testing_path=_testing_path
        )

        if camera is not None:
            self.camera_chooser._choose_existing.value = camera

        # Do not show the camera details by default
        self.camera_chooser.display_details = False

        image_camer_box = ipw.HBox()
        image_camer_box.children = [self.fits_file.file_chooser, self.camera_chooser]

        im_view_plot_box = ipw.GridspecLayout(1, 2)

        # Box for aperture settings and title
        ap_setting_box = ipw.VBox()

        self.ap_title = ipw.HTML(value=self._format_title(DEFAULT_SAVE_TITLE))
        self.aperture_settings = ui_generator(PhotometryApertures)
        self.aperture_settings.show_savebuttonbar = True
        self.aperture_settings.savebuttonbar.fns_onsave_add_action(self.save)

        ap_setting_box.children = [
            self.ap_title,
            self.aperture_settings,
        ]

        plot_box = ipw.VBox()
        plt_tabs = ipw.Tab()
        plt_tabs.children = [
            self.snr_plot,
            self.seeing_profile_plot,
            self.curve_growth_plot,
        ]
        plt_tabs.titles = [
            "SNR",
            "Seeing profile",
            "Integrated counts",
        ]

        self.tess_box = self._make_tess_box()
        plot_box.children = [plt_tabs, self.tess_box]

        imbox = ipw.VBox()
        imbox.children = [imagewidget]
        im_view_plot_box[0, 0] = imbox
        im_view_plot_box[0, 1] = plot_box
        im_view_plot_box.layout.width = "100%"

        # Line below puts space between the image and the plots so the plots
        # don't jump around as the image value changes.
        im_view_plot_box.layout.justify_content = "space-between"
        self.big_box = im_view_plot_box
        self.container.children = [
            image_camer_box,
            self.error_console,
            self.big_box,
            ap_setting_box,
        ]
        self.box = self.container
        self._aperture_name = "aperture"

        self._tess_sub = None

        # This is eventually used to store the radial profile
        self.rad_prof = None
        # Fill these in later with name of object from FITS file
        self.object_name = ""
        self.exposure = 0
        self._set_observers()
        self.aperture_settings.description = ""

    @property
    def camera(self):
        return self.camera_chooser.value

    def load_fits(self):
        """
        Load a FITS file into the image widget.
        """
        self.fits_file.load_in_image_widget(self.iw)
        self.object_name = self.fits_file.object
        for key in EXPOSURE_KEYWORDS:
            if key in self.fits_file.header:
                self.exposure = self.fits_file.header[key]
                break
        else:
            # apparently setting a higher stacklevel is better, see
            # https://docs.astral.sh/ruff/rules/no-explicit-stacklevel/
            warnings.warn(
                "No exposure time keyword found in FITS header. "
                "Setting exposure to NaN",
                stacklevel=2,
            )
            self.exposure = np.nan

    def save(self):
        """
        Save all of the settings we have to a partial settings file.
        """
        self.photometry_settings.save(
            PartialPhotometrySettings(
                photometry_apertures=self.aperture_settings.value, camera=self.camera
            ),
            update=True,
        )

        # For some reason the value of unsaved_changes is not updated until after this
        # function executes, so we force its value here.
        self.aperture_settings.savebuttonbar.unsaved_changes = False
        # Update the save box title to reflect the save
        self._set_save_box_title("")

    def _format_title(self, title):
        """
        Format titles in a consistent way.
        """
        return f"<h2>{title}</h2>"

    def _update_file(self, change):  # noqa: ARG002
        # Widget callbacks need to accept a single argument, even if it is not used.
        self.load_fits()

    def _construct_tess_sub(self):
        file = self.fits_file.path
        self._tess_sub = TessSubmission.from_header(
            fits.getheader(file),
            telescope_code=self.setting_box.telescope_code.value,
            planet=self.setting_box.planet_num.value,
        )

    def _set_seeing_profile_name(self, change):  # noqa: ARG002
        """
        Widget callbacks need to accept a single argument, even if it is not used.
        """
        self._construct_tess_sub()
        self.seeing_file_name.value = self._tess_sub.seeing_profile

    def _save_toggle_action(self, change):
        activated = change["new"]

        if activated:
            self.setting_box.layout.visibility = "visible"
            self._set_seeing_profile_name("")
        else:
            self.setting_box.layout.visibility = "hidden"

    def _save_seeing_plot(self, button):  # noqa: ARG002
        """
        Widget button callbacks need to accept a single argument.
        """
        self._seeing_plot_fig.savefig(self.seeing_file_name.value)

    def _set_save_box_title(self, change):
        # If we got here via a traitlets event then change is a dict, check that
        # case first.
        dirty = False

        try:
            if change["new"] != change["old"]:
                dirty = True
        except (KeyError, TypeError):
            dirty = False

        # The unsaved_changes attribute is not a traitlet, and it isn't clear when
        # in the event handling it gets set. When not called from an event, though,
        # this function can only used unsaved_changes to decide what the title
        # should be.
        if self.aperture_settings.savebuttonbar.unsaved_changes or dirty:
            self.ap_title.value = self._format_title(
                f"{DEFAULT_SAVE_TITLE} {AP_SETTING_NEEDS_SAVE}"
            )
        else:
            self.ap_title.value = self._format_title(
                f"{DEFAULT_SAVE_TITLE} {AP_SETTING_SAVED}"
            )

    def _set_observers(self):
        def aperture_obs(change):
            self._update_plots()
            ape = PhotometryApertures(**change["new"])
            self.aperture_settings.description = (
                f"Inner annulus: {ape.inner_annulus}, "
                f"outer annulus: {ape.outer_annulus}"
            )

        self.aperture_settings.observe(aperture_obs, names="_value")

        self.fits_file.file_chooser.observe(self._update_file, names="_value")

        self.aperture_settings.observe(self._set_save_box_title, "_value")

        if self.save_toggle:
            self.save_toggle.observe(self._save_toggle_action, names="value")
            self.save_seeing.on_click(self._save_seeing_plot)
            self.setting_box.planet_num.observe(self._set_seeing_profile_name)
            self.setting_box.telescope_code.observe(self._set_seeing_profile_name)

    def _make_tess_box(self):
        box = ipw.VBox()

        if self.observatory is None or self.observatory.TESS_telescope_code is None:
            """
            No TESS information, so definitely don't display this group of settings.
            """
            box.layout.flex_flow = "row wrap"
            box.layout.visibility = "hidden"

            self.save_toggle = None
            return box

        setting_box = ipw.HBox()
        self.save_toggle = ipw.ToggleButton(
            description="TESS seeing profile...", disabled=True
        )
        scope_name = ipw.Text(
            description="Telescope code",
            value=self.observatory.TESS_telescope_code,
            style=DESC_STYLE,
        )
        planet_num = ipw.IntText(description="Planet", value=1)
        self.save_seeing = ipw.Button(description="Save")
        self.seeing_file_name = ipw.Label(value="Moo")
        setting_box.children = (
            scope_name,
            planet_num,
            self.seeing_file_name,
            self.save_seeing,
        )
        # for kid in setting_box.children:
        #     kid.disabled = True
        box.children = (self.save_toggle, setting_box)
        setting_box.telescope_code = scope_name
        setting_box.planet_num = planet_num
        setting_box.layout.flex_flow = "row wrap"
        setting_box.layout.visibility = "hidden"
        self.setting_box = setting_box
        return box

    def _update_ap_settings(self, value):
        self.aperture_settings.value = value

    def _make_show_event(self):
        def show_event(
            viewer, event=None, datax=None, datay=None, aperture=None  # noqa: ARG001
        ):
            """
            ginga callbacks require the function signature above.
            """
            profile_size = 60
            centering_cutout_size = 20
            default_gap = 5  # pixels
            default_annulus_width = 15  # pixels
            if self.save_toggle:
                self.save_toggle.disabled = False

            update_aperture_settings = False
            if event is not None:
                # User clicked on a star, so generate profile
                i = self.iw._viewer.get_image()
                data = i.get_data()

                # Rough location of click in original image
                x = int(np.floor(event.data_x))
                y = int(np.floor(event.data_y))

                try:
                    rad_prof = CenterAndProfile(
                        data,
                        (x, y),
                        profile_radius=profile_size,
                        centering_cutout_size=centering_cutout_size,
                    )
                except RuntimeError as e:
                    # Check whether this error is one generated by RadialProfile
                    if "Centroid did not converge on a star." in str(e):
                        # Clear any previous messages...no idea why the clear_output
                        # method doesn't work here, but it doesn't/
                        self.error_console.outputs = ()

                        # Use the append_display_data method instead of the
                        # error_console context manager because there seems to be
                        # a timing issue with the context manager when running
                        # tests.
                        self.error_console.append_display_data(
                            ipw.HTML(
                                "<strong>No star found at this location. "
                                "Try clicking closer "
                                "to a star or on a brighter star</strong>"
                            )
                        )
                        print(f"{self.error_console.outputs=}")
                        return
                    else:
                        # RadialProfile did not generate this error, pass it
                        # on to the user
                        raise e  # pragma: no cover
                else:
                    # Success, clear any previous error messages
                    self.error_console.clear_output()

                try:
                    try:  # Remove previous marker
                        self.iw.remove_markers(marker_name=self._aperture_name)
                    except AttributeError:
                        self.iw.remove_markers_by_name(marker_name=self._aperture_name)
                except ValueError:
                    # No markers yet, keep going
                    pass

                # ADD MARKER WHERE CLICKED
                self.iw.add_markers(
                    Table(
                        data=[[rad_prof.center[0]], [rad_prof.center[1]]],
                        names=["x", "y"],
                    ),
                    marker_name=self._aperture_name,
                )

                # Default is 1.5 times FWHM
                aperture_radius = np.round(1.5 * rad_prof.FWHM, 0)
                self.rad_prof = rad_prof

                # Make an aperture settings object, but don't update it's widget yet.
                ap_settings = PhotometryApertures(
                    radius=aperture_radius,
                    gap=default_gap,
                    annulus_width=default_annulus_width,
                    fwhm=rad_prof.FWHM,
                )
                update_aperture_settings = True
            else:
                # User changed aperture
                aperture_radius = aperture["radius"]
                ap_settings = PhotometryApertures(
                    **aperture
                )  # Make an aperture settings object, but don't update it's widget yet.

            if update_aperture_settings:
                # So it turns out that the validation stuff only updates when changes
                # are made in the UI rather than programmatically. Since we know we've
                # set a valid value, and that we've made changes we just manually set
                # the relevant values.
                self.aperture_settings.savebuttonbar.unsaved_changes = True
                self.aperture_settings.is_valid.value = True

                # Update the value last so that the unsaved state is properly set when
                # the value is updated.
                self._update_ap_settings(ap_settings.model_dump())

            self._update_plots()

        return show_event

    def _update_plots(self):
        # DISPLAY THE SCALED PROFILE
        fig_size = (10, 5)

        # Stop if the update is happening before a radial profile has been generated
        # (e.g. the user changes the aperture settings before loading an image).
        if self.rad_prof is None:
            return

        rad_prof = self.rad_prof
        self.seeing_profile_plot.clear_output(wait=True)
        ap_settings = PhotometryApertures(**self.aperture_settings.value)
        with self.seeing_profile_plot:
            r_exact, individual_counts = rad_prof.pixel_values_in_profile
            scaled_exact_counts = (
                individual_counts / rad_prof.radial_profile.profile.max()
            )
            self._seeing_plot_fig = seeing_plot(
                r_exact,
                scaled_exact_counts,
                rad_prof.radial_profile.radius,
                rad_prof.normalized_profile,
                rad_prof.HWHM,
                self.object_name,
                photometry_settings=ap_settings,
                figsize=fig_size,
            )
            plt.show()

        # CALCULATE AND DISPLAY NET COUNTS INSIDE RADIUS
        self.curve_growth_plot.clear_output(wait=True)
        with self.curve_growth_plot:
            cog = rad_prof.curve_of_growth

            plt.figure(figsize=fig_size)
            plt.plot(cog.radius, cog.profile)
            plt.xlim(0, 40)
            ylim = plt.ylim()
            plt.vlines(ap_settings.radius, *plt.ylim(), colors=["red"])
            plt.ylim(*ylim)
            plt.grid()

            plt.title("Net counts in aperture")

            plt.xlabel("Aperture radius (pixels)")
            plt.ylabel(f"Net counts ({self.camera.data_unit})")
            plt.show()

        # CALCULATE And DISPLAY SNR AS A FUNCTION OF RADIUS
        self.snr_plot.clear_output(wait=True)
        with self.snr_plot:
            plt.figure(figsize=fig_size)
            snr = rad_prof.snr(self.camera, self.exposure)

            plt.plot(rad_prof.curve_of_growth.radius, snr)

            plt.title(
                f"Signal to noise ratio max {snr.max():.1f} "
                f"at radius {snr.argmax() + 1}"
            )
            plt.xlim(0, 40)
            ylim = plt.ylim()
            plt.vlines(ap_settings.radius, *plt.ylim(), colors=["red"])
            plt.ylim(*ylim)
            plt.xlabel("Aperture radius (pixels)")
            plt.ylabel("SNR")
            plt.grid()
            plt.show()
